from typing import Any, Literal, TypeAlias

from typing_extensions import Self

from ..context import context
from ..element import Element
from ..events import Handler, UiEventArguments, handle_event

NotificationPosition = Literal[
    'top-left',
    'top-right',
    'bottom-left',
    'bottom-right',
    'top',
    'bottom',
    'left',
    'right',
    'center',
]

NotificationType: TypeAlias = Literal[
    'positive',
    'negative',
    'warning',
    'info',
    'ongoing',
] | None


class Notification(Element, component='notification.js'):

    def __init__(self,
                 message: Any = '', *,
                 position: NotificationPosition = 'bottom',
                 close_button: bool | str = False,
                 type: NotificationType = None,  # pylint: disable=redefined-builtin
                 color: str | None = None,
                 multi_line: bool = False,
                 icon: str | None = None,
                 spinner: bool = False,
                 timeout: float | None = 5.0,
                 on_dismiss: Handler[UiEventArguments] | None = None,
                 options: dict | None = None,
                 **kwargs: Any,
                 ) -> None:
        """Notification element

        Displays a notification on the screen.
        In contrast to `ui.notify`, this element allows to update the notification message and other properties once the notification is displayed.
        The notification can be removed with `dismiss()`.

        :param message: content of the notification
        :param position: position on the screen ("top-left", "top-right", "bottom-left", "bottom-right", "top", "bottom", "left", "right" or "center", default: "bottom")
        :param close_button: optional label of a button to dismiss the notification (default: `False`)
        :param type: optional type ("positive", "negative", "warning", "info" or "ongoing")
        :param color: optional color name
        :param multi_line: enable multi-line notifications
        :param icon: optional name of an icon to be displayed in the notification (default: `None`)
        :param spinner: display a spinner in the notification (default: False)
        :param timeout: optional timeout in seconds after which the notification is dismissed (default: 5.0)
        :param on_dismiss: optional callback to be invoked when the notification is dismissed
        :param options: optional dictionary with all options (overrides all other arguments)

        Note: You can pass additional keyword arguments according to `Quasar's Notify API <https://quasar.dev/quasar-plugins/notify#notify-api>`_.
        """
        with context.client.layout:
            super().__init__()
        self._update_method = 'update_notification'
        if options:
            self._props['options'] = options
        else:
            self._props['options'] = {
                'message': str(message),
                'position': position,
                'multiLine': multi_line,
                'spinner': spinner,
                'closeBtn': close_button,
                'timeout': (timeout or 0) * 1000,
                'group': False,
            }
            if type is not None:
                self._props['options']['type'] = type
            if color is not None:
                self._props['options']['color'] = color
            if icon is not None:
                self._props['options']['icon'] = icon
            self._props['options'].update(kwargs)

        if on_dismiss:
            self.on_dismiss(on_dismiss)

        async def handle_dismiss() -> None:
            if not self._deleted:
                self.clear()
                self.delete()
        self.on('dismiss', handle_dismiss)

    @property
    def message(self) -> str:
        """Message text."""
        return self._props['options']['message']

    @message.setter
    def message(self, value: Any) -> None:
        self._props['options']['message'] = str(value)

    @property
    def position(self) -> NotificationPosition:
        """Position on the screen."""
        return self._props['options']['position']

    @position.setter
    def position(self, value: NotificationPosition) -> None:
        self._props['options']['position'] = value

    @property
    def type(self) -> NotificationType:
        """Type of the notification."""
        return self._props['options'].get('type')

    @type.setter
    def type(self, value: NotificationType) -> None:
        if value is None:
            self._props['options'].pop('type', None)
        else:
            self._props['options']['type'] = value

    @property
    def color(self) -> str | None:
        """Color of the notification."""
        return self._props['options'].get('color')

    @color.setter
    def color(self, value: str | None) -> None:
        if value is None:
            self._props['options'].pop('color', None)
        else:
            self._props['options']['color'] = value

    @property
    def multi_line(self) -> bool:
        """Whether the notification is multi-line."""
        return self._props['options']['multiLine']

    @multi_line.setter
    def multi_line(self, value: bool) -> None:
        self._props['options']['multiLine'] = value

    @property
    def icon(self) -> str | None:
        """Icon of the notification."""
        return self._props['options'].get('icon')

    @icon.setter
    def icon(self, value: str | None) -> None:
        if value is None:
            self._props['options'].pop('icon', None)
        else:
            self._props['options']['icon'] = value

    @property
    def spinner(self) -> bool:
        """Whether the notification is a spinner."""
        return self._props['options']['spinner']

    @spinner.setter
    def spinner(self, value: bool) -> None:
        self._props['options']['spinner'] = value

    @property
    def timeout(self) -> float:
        """Timeout of the notification in seconds.

        *Added in version 2.13.0*
        """
        return self._props['options']['timeout'] / 1000

    @timeout.setter
    def timeout(self, value: float | None) -> None:
        self._props['options']['timeout'] = (value or 0) * 1000

    @property
    def close_button(self) -> bool | str:
        """Whether the notification has a close button."""
        return self._props['options']['closeBtn']

    @close_button.setter
    def close_button(self, value: bool | str) -> None:
        self._props['options']['closeBtn'] = value

    def on_dismiss(self, callback: Handler[UiEventArguments]) -> Self:
        """Add a callback to be invoked when the notification is dismissed."""
        self.on('dismiss', lambda _: handle_event(callback, UiEventArguments(sender=self, client=self.client)), [])
        return self

    def dismiss(self) -> None:
        """Dismiss the notification."""
        self.run_method('dismiss')

    def set_visibility(self, visible: bool) -> None:
        raise NotImplementedError('Use `dismiss()` to remove the notification. See #3670 for more information.')



from collections.abc import Callable, Generator, Iterable, Iterator
from contextlib import suppress
from copy import deepcopy
from typing import Any, Literal

from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import GenericEventArguments, Handler, ValueChangeEventArguments
from .choice_element import ChoiceElement
from .mixins.disableable_element import DisableableElement
from .mixins.label_element import LabelElement
from .mixins.validation_element import ValidationDict, ValidationElement, ValidationFunction


class Select(LabelElement, ValidationElement, ChoiceElement, DisableableElement, component='select.js'):

    @resolve_defaults
    def __init__(self,
                 options: list | dict, *,
                 label: str | None = DEFAULT_PROP | None,
                 value: Any = DEFAULT_PROPS['model-value'] | None,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 with_input: bool = False,
                 new_value_mode: Literal['add', 'add-unique', 'toggle'] | None = DEFAULT_PROP | None,
                 multiple: bool = DEFAULT_PROP | False,
                 clearable: bool = DEFAULT_PROP | False,
                 validation: ValidationFunction | ValidationDict | None = None,
                 key_generator: Callable[[Any], Any] | Iterator[Any] | None = None,
                 ) -> None:
        """Dropdown Selection

        This element is based on Quasar's `QSelect <https://quasar.dev/vue-components/select>`_ component.

        The options can be specified as a list of values, or as a dictionary mapping values to labels.
        After manipulating the options, call `update()` to update the options in the UI.

        If `with_input` is True, an input field is shown to filter the options.

        If `new_value_mode` is not None, it implies `with_input=True` and the user can enter new values in the input field.
        See `Quasar's documentation <https://quasar.dev/vue-components/select#the-new-value-mode-prop>`_ for details.
        Note that this mode is ineffective when setting the `value` property programmatically.

        You can use the `validation` parameter to define a dictionary of validation rules,
        e.g. ``{'Too long!': lambda value: len(value) < 3}``.
        The key of the first rule that fails will be displayed as an error message.
        Alternatively, you can pass a callable that returns an optional error message.
        To disable the automatic validation on every value change, you can use the `without_auto_validation` method.

        :param options: a list ['value1', ...] or dictionary `{'value1':'label1', ...}` specifying the options
        :param label: the label to display above the selection
        :param value: the initial value
        :param on_change: callback to execute when selection changes
        :param with_input: whether to show an input field to filter the options
        :param new_value_mode: handle new values from user input (default: None, i.e. no new values)
        :param multiple: whether to allow multiple selections
        :param clearable: whether to add a button to clear the selection
        :param validation: dictionary of validation rules or a callable that returns an optional error message (default: None for no validation)
        :param key_generator: a callback or iterator to generate a dictionary key for new values
        """
        self.multiple = multiple
        if multiple:
            if value is None:
                value = []
            elif not isinstance(value, list):
                value = [value]
            else:
                value = value[:]  # NOTE: avoid modifying the original list which could be the list of options (#3014)
        super().__init__(label=label, options=options, value=value, on_change=on_change, validation=validation)
        if isinstance(key_generator, Generator):
            next(key_generator)  # prime the key generator, prepare it to receive the first value
        self.key_generator = key_generator
        if new_value_mode is not None:
            if isinstance(options, dict) and new_value_mode == 'add' and key_generator is None:
                raise ValueError('new_value_mode "add" is not supported for dict options without key_generator')
            self._props['new-value-mode'] = new_value_mode
            with_input = True
        if with_input:
            self.original_options = deepcopy(options)
            self._props['use-input'] = True
            self._props.set_bool('hide-selected', not multiple)
            self._props['fill-input'] = True
            self._props['input-debounce'] = 0
        self._props.set_bool('multiple', multiple)
        self._props.set_bool('clearable', clearable)

        self._is_showing_popup = False
        self.on('popup-show', lambda e: setattr(e.sender, '_is_showing_popup', True))
        self.on('popup-hide', lambda e: setattr(e.sender, '_is_showing_popup', False))

    @property
    def is_showing_popup(self) -> bool:
        """Whether the options popup is currently shown."""
        return self._is_showing_popup

    def _event_args_to_value(self, e: GenericEventArguments) -> Any:
        # pylint: disable=too-many-nested-blocks
        if self.multiple:
            if e.args is None:
                return []
            else:
                if self._props.get('new-value-mode') == 'add-unique':
                    # handle issue #4896: eliminate duplicate arguments
                    for arg1 in [a for a in e.args if isinstance(a, str)]:
                        for arg2 in [a for a in e.args if isinstance(a, dict)]:
                            if arg1 == arg2['label']:
                                e.args.remove(arg1)
                                break
                args = [self._values[arg['value']] if isinstance(arg, dict) else arg for arg in e.args]
                for arg in e.args:
                    if isinstance(arg, str):
                        self._handle_new_value(arg)
                return [arg for arg in args if arg in self._values]
        else:  # noqa: PLR5501
            if e.args is None:
                return None
            else:  # noqa: PLR5501
                if isinstance(e.args, str):
                    new_value = self._handle_new_value(e.args)
                    return new_value if new_value in self._values else None
                else:
                    return self._values[e.args['value']]

    def _value_to_model_value(self, value: Any) -> Any:
        # pylint: disable=no-else-return
        if self.multiple:
            result = []
            for item in value or []:
                with suppress(ValueError):
                    index = self._values.index(item)
                    result.append({'value': index, 'label': self._labels[index]})
            return result
        else:
            try:
                index = self._values.index(value)
                return {'value': index, 'label': self._labels[index]}
            except ValueError:
                return None

    def _generate_key(self, value: str) -> Any:
        if isinstance(self.key_generator, Generator):
            return self.key_generator.send(value)
        if isinstance(self.key_generator, Iterable):
            return next(self.key_generator)
        if callable(self.key_generator):
            return self.key_generator(value)
        return value

    def _handle_new_value(self, value: str) -> Any:
        mode = self._props['new-value-mode']
        if isinstance(self.options, list):
            if mode == 'add':
                self.options.append(value)
            elif mode == 'add-unique':
                if value not in self.options:
                    self.options.append(value)
            elif mode == 'toggle':
                if value in self.options:
                    self.options.remove(value)
                else:
                    self.options.append(value)
            # NOTE: self._labels and self._values are updated via self.options since they share the same references
            return value
        else:
            key = value
            if mode == 'add':
                key = self._generate_key(value)
                self.options[key] = value
            elif mode == 'add-unique':
                if value not in self.options.values():
                    key = self._generate_key(value)
                    self.options[key] = value
            elif mode == 'toggle':
                if value in self.options:
                    self.options.pop(value)
                else:
                    key = self._generate_key(value)
                    self.options.update({key: value})
            self._update_values_and_labels()
            return key



from collections.abc import Callable
from pathlib import Path

import fastapi

from .page import page as ui_page


class APIRouter(fastapi.APIRouter):

    def page(self,
             path: str, *,
             title: str | None = None,
             viewport: str | None = None,
             favicon: str | Path | None = None,
             dark: bool | None = ...,  # type: ignore
             response_timeout: float = 3.0,
             **kwargs,
             ) -> Callable:
        """Page

        Creates a new page at the given route.
        Each user will see a new instance of the page.
        This means it is private to the user and not shared with others.

        :param path: route of the new page (path must start with '/')
        :param title: optional page title
        :param viewport: optional viewport meta tag content
        :param favicon: optional relative filepath or absolute URL to a favicon (default: `None`, NiceGUI icon will be used)
        :param dark: whether to use Quasar's dark mode (defaults to `dark` argument of `run` command)
        :param response_timeout: maximum time for the decorated function to build the page (default: 3.0)
        :param kwargs: additional keyword arguments passed to FastAPI's @app.get method
        """
        return ui_page(
            path,
            title=title,
            viewport=viewport,
            favicon=favicon,
            dark=dark,
            response_timeout=response_timeout,
            api_router=self,
            **kwargs
        )



import os
import runpy
from collections.abc import AsyncGenerator, Callable
from contextlib import asynccontextmanager
from pathlib import Path

import httpx

from .. import core, ui
from ..functions.download import download
from ..functions.navigate import Navigate
from ..functions.notify import notify
from .general import nicegui_reset_globals, prepare_simulation
from .user import User


@asynccontextmanager
async def user_simulation(
    root: Callable | None = None, *, main_file: str | os.PathLike | None = None,
) -> AsyncGenerator[User]:
    """Context manager for test user simulation.

    This context manager yields a ``User`` connected to a NiceGUI app within an isolated test context.

    :param root: root function which is passed directly to ``ui.run``; mutually exclusive with ``main_file`` argument.
    :param main_file: path to a NiceGUI main file executed via ``runpy.run_path``; mutually exclusive with ``root`` argument.
    """
    if main_file is not None and root is not None:
        raise ValueError('Cannot specify both `main_file` and `root` function simultaneously.')

    with nicegui_reset_globals():
        os.environ['NICEGUI_USER_SIMULATION'] = 'true'
        try:
            if main_file is not None:
                if not Path(main_file).exists():
                    raise FileNotFoundError(f'Main file not found at {main_file}')
                runpy.run_path(str(main_file), run_name='__main__')
            else:
                prepare_simulation()
                ui.run(root, storage_secret='simulated secret')

            async with core.app.router.lifespan_context(core.app):
                async with httpx.AsyncClient(transport=httpx.ASGITransport(core.app), base_url='http://test') as client:
                    yield User(client)
        finally:
            os.environ.pop('NICEGUI_USER_SIMULATION', None)
            ui.navigate = Navigate()
            ui.notify = notify
            ui.download = download



from typing import Literal

from typing_extensions import Self

from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import ClickEventArguments, Handler, handle_event
from .mixins.color_elements import BackgroundColorElement
from .mixins.disableable_element import DisableableElement
from .mixins.icon_element import IconElement
from .mixins.label_element import LabelElement
from .mixins.value_element import ValueElement


class Fab(ValueElement, LabelElement, IconElement, BackgroundColorElement, DisableableElement):

    @resolve_defaults
    def __init__(self,
                 icon: str, *,
                 value: bool = DEFAULT_PROPS['model-value'] | False,
                 label: str = DEFAULT_PROP | '',
                 color: str | None = DEFAULT_PROP | 'primary',
                 direction: Literal['up', 'down', 'left', 'right'] = DEFAULT_PROP | 'right',
                 ) -> None:
        """Floating Action Button (FAB)

        A floating action button that can be used to trigger an action.
        This element is based on Quasar's `QFab <https://quasar.dev/vue-components/floating-action-button#qfab-api>`_ component.

        :param icon: icon to be displayed on the FAB
        :param value: whether the FAB is already opened (default: ``False``)
        :param label: optional label for the FAB
        :param color: background color of the FAB (default: "primary")
        :param direction: direction of the FAB ("up", "down", "left", "right", default: "right")
        """
        super().__init__(tag='q-fab', value=value, label=label, background_color=color, icon=icon)
        self._props['direction'] = direction

    def open(self) -> None:
        """Open the FAB."""
        self.value = True

    def close(self) -> None:
        """Close the FAB."""
        self.value = False

    def toggle(self) -> None:
        """Toggle the FAB."""
        self.value = not self.value


class FabAction(LabelElement, IconElement, BackgroundColorElement, DisableableElement):

    @resolve_defaults
    def __init__(self,
                 icon: str, *,
                 label: str = DEFAULT_PROP | '',
                 on_click: Handler[ClickEventArguments] | None = None,
                 color: str | None = DEFAULT_PROP | 'primary',
                 auto_close: bool = True,
                 ) -> None:
        """Floating Action Button Action

        An action that can be added to a floating action button (FAB).
        This element is based on Quasar's `QFabAction <https://quasar.dev/vue-components/floating-action-button#qfabaction-api>`_ component.

        *Added in version 2.22.0*

        :param icon: icon to be displayed on the action button
        :param label: optional label for the action button
        :param color: background color of the action button (default: "primary")
        :param auto_close: whether the FAB should be closed after a click event (default: ``True``)
        """
        super().__init__(tag='q-fab-action', label=label, background_color=color, icon=icon)
        self.fab = next((e for e in self.ancestors() if isinstance(e, Fab)), None)
        if self.fab and not auto_close:
            self.on('click', self.fab.open)

        if on_click:
            self.on_click(on_click)

    def on_click(self, callback: Handler[ClickEventArguments]) -> Self:
        """Add a callback to be invoked when the action element is clicked."""
        self.on('click', lambda _: handle_event(callback, ClickEventArguments(sender=self, client=self.client)), [])
        return self



from pathlib import Path

from ..defaults import DEFAULT_PROP, resolve_defaults
from .mixins.source_element import SourceElement


class Video(SourceElement, component='video.js'):
    SOURCE_IS_MEDIA_FILE = True

    @resolve_defaults
    def __init__(self, src: str | Path, *,
                 controls: bool = DEFAULT_PROP | True,
                 autoplay: bool = DEFAULT_PROP | False,
                 muted: bool = DEFAULT_PROP | False,
                 loop: bool = DEFAULT_PROP | False,
                 ) -> None:
        """Video

        Displays a video.

        :param src: URL or local file path of the video source
        :param controls: whether to show the video controls, like play, pause, and volume (default: `True`)
        :param autoplay: whether to start playing the video automatically (default: `False`)
        :param muted: whether the video should be initially muted (default: `False`)
        :param loop: whether the video should loop (default: `False`)

        See `here <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/video#events>`_
        for a list of events you can subscribe to using the generic event subscription `on()`.
        """
        super().__init__(source=src)
        self._props.set_bool('controls', controls)
        self._props.set_bool('autoplay', autoplay)
        self._props.set_bool('muted', muted)
        self._props.set_bool('loop', loop)

    def set_source(self, source: str | Path) -> None:
        return super().set_source(source)

    def seek(self, seconds: float) -> None:
        """Seek to a specific position in the video.

        :param seconds: the position in seconds
        """
        self.run_method('seek', seconds)

    def play(self) -> None:
        """Play video."""
        self.run_method('play')

    def pause(self) -> None:
        """Pause video."""
        self.run_method('pause')



from __future__ import annotations

from typing import TYPE_CHECKING

from . import core
from .slot import Slot

if TYPE_CHECKING:
    from .client import Client


class Context:

    @property
    def slot_stack(self) -> list[Slot]:
        """Return the slot stack of the current asyncio task."""
        stack = Slot.get_stack()
        if not stack and not core.script_mode and not core.app.is_started:
            # create a pseudo client to "survive" until reaching `ui.run()`
            from .client import Client  # pylint: disable=import-outside-toplevel,cyclic-import
            from .page import page  # pylint: disable=import-outside-toplevel,cyclic-import
            if not Client.instances:  # in case some kind of dummy client is already created
                core.script_mode = True
                core.script_client = Client(page('/')).__enter__()  # pylint: disable=unnecessary-dunder-call
                stack = Slot.get_stack()
        return stack

    @property
    def slot(self) -> Slot:
        """Return the current slot."""
        slot_stack = self.slot_stack
        if not slot_stack:
            raise RuntimeError('The current slot cannot be determined because the slot stack for this task is empty.\n'
                               'This may happen if you try to create UI from a background task.\n'
                               'To fix this, enter the target slot explicitly using `with container_element:`.')
        return slot_stack[-1]

    @property
    def client(self) -> Client:
        """Return the current client."""
        return self.slot.parent.client


context = Context()



import html
from collections.abc import Callable

from .. import helpers
from ..defaults import DEFAULT_PROP, resolve_defaults
from .html import Html
from .mixins.label_element import LabelElement


class ChatMessage(LabelElement):

    @resolve_defaults
    def __init__(self,
                 text: str | list[str] | None = None,
                 name: str | None = DEFAULT_PROP | None,
                 label: str | None = DEFAULT_PROP | None,
                 stamp: str | None = DEFAULT_PROP | None,
                 avatar: str | None = DEFAULT_PROP | None,
                 sent: bool = DEFAULT_PROP | False,
                 text_html: bool = False,
                 sanitize: Callable[[str], str] | bool | None = True,  # DEPRECATED: remove `None` in version 4.0.0
                 ) -> None:
        """Chat Message

        Based on Quasar's `Chat Message <https://quasar.dev/vue-components/chat/>`_ component.

        :param text: the message body (can be a list of strings for multiple message parts)
        :param name: the name of the message author
        :param label: renders a label header/section only
        :param stamp: timestamp of the message
        :param avatar: URL to an avatar
        :param sent: render as a sent message (so from current user) (default: ``False``)
        :param text_html: render text as HTML (default: ``False``)
        :param sanitize: sanitization mode (only relevant when ``text_html=True``):
            ``True`` (default) uses client-side sanitization via DOMPurify,
            ``False`` disables sanitization (use only with trusted content),
            or pass a callable to apply server-side sanitization
        """
        super().__init__(tag='q-chat-message', label=label)

        if sanitize is None:
            helpers.warn_once('`sanitize=None` is deprecated, defaults to `True` and will be removed in version 4.0.0.')
            sanitize = True  # DEPRECATED: remove this block in version 4.0.0

        if text is None:
            text = []
        if isinstance(text, str):
            text = [text]
        if not text_html:
            text = [html.escape(part) for part in text]
            text = [part.replace('\n', '<br />') for part in text]
            sanitize = False

        self._props.set_optional('name', name)
        self._props.set_optional('stamp', stamp)
        self._props.set_optional('avatar', avatar)
        self._props['sent'] = sent

        with self:
            for line in text:
                Html(line, sanitize=sanitize)



import importlib.util
import json
from datetime import date, datetime
from typing import Any

from fastapi.responses import JSONResponse

try:
    HAS_NUMPY = importlib.util.find_spec('numpy') is not None
except (ModuleNotFoundError, ValueError):
    HAS_NUMPY = False


def dumps(obj: Any,
          sort_keys: bool = False,
          separators: tuple[str, str] | None = None, *,
          indent: bool = False) -> str:
    """Serializes a Python object to a JSON-encoded string.

    This implementation uses Python's default json module, but extends it in order to support NumPy arrays.
    """
    if separators is None:
        separators = (',', ':')
    return json.dumps(
        obj,
        sort_keys=sort_keys,
        separators=separators,
        indent=2 if indent else None,
        allow_nan=False,
        ensure_ascii=False,
        cls=NumpyJsonEncoder)


def loads(value: str) -> Any:
    """Deserialize a JSON-encoded string to a corresponding Python object/value.

    Uses Python's default json module internally.
    """
    return json.loads(value)


class NiceGUIJSONResponse(JSONResponse):
    """FastAPI response class to support our custom json serializer implementation."""

    def render(self, content: Any) -> bytes:
        return dumps(content).encode('utf-8')


class NumpyJsonEncoder(json.JSONEncoder):
    """Special json encoder that supports NumPy arrays and date/datetime objects."""

    def default(self, o):
        if HAS_NUMPY:
            import numpy as np  # pylint: disable=import-outside-toplevel
            if isinstance(o, np.integer):
                return int(o)
            if isinstance(o, np.floating):
                return float(o)
            if isinstance(o, np.ndarray):
                return o.tolist()
        if isinstance(o, (datetime, date)):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)



import weakref
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path, PurePosixPath

import aiofiles
import anyio
from starlette.datastructures import UploadFile
from starlette.formparsers import MultiPartParser

from .. import json, run


@dataclass
class FileUpload(ABC):
    name: str
    content_type: str

    @abstractmethod
    async def read(self) -> bytes:
        """Read the file contents as bytes."""

    @abstractmethod
    async def text(self, encoding: str = 'utf-8') -> str:
        """Read the file contents as text.

        :param encoding: the encoding to use for the text (default: "utf-8")
        """

    async def json(self, encoding: str = 'utf-8') -> dict:
        """Read the file contents as JSON dictionary.

        :param encoding: the encoding to use for the text (default: "utf-8")
        """
        return json.loads(await self.text(encoding))

    @abstractmethod
    def iterate(self, *, chunk_size: int = 1024 * 1024) -> AsyncIterator[bytes]:
        """Iterate over the file contents as bytes.

        :param chunk_size: the size of each chunk to read in bytes (default: 1 MB)
        """

    @abstractmethod
    async def save(self, path: str | Path) -> None:
        """Save the file contents to a path.

        :param path: the path to save the file contents to
        """

    @abstractmethod
    def size(self) -> int:
        """Get the file size in bytes."""


@dataclass
class SmallFileUpload(FileUpload):
    _data: bytes

    def size(self) -> int:
        return len(self._data)

    async def read(self) -> bytes:
        return self._data

    async def text(self, encoding: str = 'utf-8') -> str:
        return self._data.decode(encoding)

    def iterate(self, *, chunk_size: int = 1024 * 1024) -> AsyncIterator[bytes]:
        async def generator() -> AsyncIterator[bytes]:
            for i in range(0, len(self._data), chunk_size):
                yield self._data[i:i + chunk_size]
        return generator()

    async def save(self, path: str | Path) -> None:
        target = Path(path)
        target.parent.mkdir(parents=True, exist_ok=True)
        await run.io_bound(target.write_bytes, self._data)


@dataclass
class LargeFileUpload(FileUpload):
    _path: Path

    def __post_init__(self) -> None:
        self._finalizer = weakref.finalize(self, _cleanup_path, self._path)

    def size(self) -> int:
        return self._path.stat().st_size

    async def read(self) -> bytes:
        async with await anyio.open_file(self._path, 'rb') as f:
            return await f.read()

    async def text(self, encoding: str = 'utf-8') -> str:
        data = await self.read()
        return data.decode(encoding)

    def iterate(self, *, chunk_size: int = 1024 * 1024) -> AsyncIterator[bytes]:
        async def generator() -> AsyncIterator[bytes]:
            async with await anyio.open_file(self._path, 'rb') as f:
                while (chunk := await f.read(chunk_size)):
                    yield chunk
        return generator()

    async def save(self, path: str | Path) -> None:
        target = Path(path)
        target.parent.mkdir(parents=True, exist_ok=True)
        async with await anyio.open_file(target, 'wb') as f:
            async for chunk in self.iterate():
                await f.write(chunk)


def _cleanup_path(path: Path) -> None:
    path.unlink(missing_ok=True)


async def create_file_upload(upload: UploadFile, *, chunk_size: int = 1024 * 1024) -> FileUpload:
    """Create a file upload from a Starlette UploadFile.

    :param upload: the Starlette UploadFile to create a file upload from
    :param chunk_size: the size of each chunk to read in bytes (default: 1 MB)
    """
    memory_limit = (
        getattr(MultiPartParser, 'spool_max_size', 0) or
        getattr(MultiPartParser, 'max_part_size', 0) or  # NOTE: for starlette < 0.46.0
        1024 * 1024
    )

    buffer = BytesIO()
    buffer_size = 0
    temp_file: aiofiles.threadpool.binary.AsyncBufferedIOBase | None = None

    try:
        while (chunk := await upload.read(chunk_size)):
            if not temp_file and buffer_size + len(chunk) > memory_limit:
                temp_file = await aiofiles.tempfile.NamedTemporaryFile('wb', delete=False)
                await temp_file.write(buffer.getvalue())
                buffer = BytesIO()  # release memory
            if not temp_file:
                buffer.write(chunk)
                buffer_size += len(chunk)
            else:
                await temp_file.write(chunk)
    finally:
        await upload.close()
        if temp_file:
            await temp_file.close()

    filename = PurePosixPath(upload.filename or '').name  # strips all path components
    if temp_file:
        return LargeFileUpload(filename, upload.content_type or '', Path(str(temp_file.name)))
    else:
        return SmallFileUpload(filename, upload.content_type or '', buffer.getvalue())



from ..element import Element


class ContextMenu(Element):

    def __init__(self) -> None:
        """Context Menu

        Creates a context menu based on Quasar's `QMenu <https://quasar.dev/vue-components/menu>`_ component.
        The context menu should be placed inside the element where it should be shown.
        It is automatically opened when the user right-clicks on the element and appears at the mouse position.
        """
        super().__init__('q-menu')
        self._props['context-menu'] = True
        self._props['touch-position'] = True

    def open(self) -> None:
        """Open the context menu."""
        self.run_method('show')

    def close(self) -> None:
        """Close the context menu."""
        self.run_method('hide')



from .mixins.text_element import TextElement


class Tooltip(TextElement):

    def __init__(self, text: str = '') -> None:
        """Tooltip

        This element is based on Quasar's `QTooltip <https://quasar.dev/vue-components/tooltip>`_ component.
        It can be placed in another element to show additional information on hover.

        Instead of passing a string as the first argument, you can also nest other elements inside the tooltip.

        :param text: the content of the tooltip (default: '')
        """
        super().__init__(tag='q-tooltip', text=text)



#!/usr/bin/env python
import argparse
import os
import platform
import subprocess
from contextlib import suppress
from pathlib import Path

import nicegui

DESCRIPTION = '''
Build a package of your NiceGUI app
-----------------------------------

NiceGUI apps can be bundled into an executable with PyInstaller.
This allows you to distribute your app as a single file that can be executed on any computer.
Use this script as a starting point to create a package for your app.

Important: Make sure to run your main script with

    ui.run(reload=False, port=native.find_open_port(), ...)

to disable the reload server and to automatically find an open port.

For more information and packaging tips, have a look into the NiceGUI documentation:
https://nicegui.io/documentation/section_configuration_deployment#package_for_installation.
'''.strip()


def main() -> None:
    parser = argparse.ArgumentParser(description=DESCRIPTION, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--name', type=str, default='Your App Name', help='Name of your app.')
    parser.add_argument('--windowed', action='store_true', default=False, help=(
        'Prevent a terminal console from appearing.\n'
        'Only use with `ui.run(native=True, ...)`.\n'
        'It will create an `.app` file on Mac which runs without showing any console output.'
    ))
    parser.add_argument('--onefile', action='store_true', default=False, help=(
        'Create a single executable file.\n'
        'Whilst convenient for distribution, it will be slower to start up.'
    ))
    parser.add_argument('--onedir', action='store_true', default=False, help=(
        'Create an executable with all supporting files in a directory.\n'
        'This starts faster than "--onefile" because it skips the unpacking step.\n'
        'For distribution, package the directory into an archive file (e.g., .zip or .7z).'
    ))
    parser.add_argument('--add-data', type=str, action='append', default=[
        f'{Path(nicegui.__file__).parent}{os.pathsep}nicegui',
    ], help='Include additional data.')
    parser.add_argument('--icon', type=str, help='Path to an icon file.')
    parser.add_argument('--osx-bundle-identifier', type=str, help='Mac OS .app bundle identifier.')
    parser.add_argument('--dry-run', action='store_true', help='Dry run.', default=False)
    parser.add_argument('--clean', action='store_true', default=False, help=(
        'Clean PyInstaller cache (in ./build folder) and remove temporary files before building.'
    ))
    parser.add_argument('--noconfirm', action='store_true', default=False, help=(
        'Replace output directory (./dist/SPECNAME) without asking for confirmation.'
    ))
    parser.add_argument('main', default='main.py', help='Main file which calls `ui.run()`.')
    args = parser.parse_args()

    command = ['pyinstaller'] if platform.system() == 'Windows' else ['python', '-m', 'PyInstaller']
    command.extend(['--name', args.name])
    if args.clean:
        command.append('--clean')
    if args.noconfirm:
        command.append('--noconfirm')
    if args.windowed:
        command.append('--windowed')
    if args.onefile:
        command.append('--onefile')
    if args.onedir:
        command.append('--onedir')
    for data in args.add_data:
        command.extend(['--add-data', data])
    if args.icon:
        command.extend(['--icon', args.icon])
    if args.osx_bundle_identifier:
        command.extend(['--osx-bundle-identifier', args.osx_bundle_identifier])

    with suppress(ImportError):
        import pyecharts  # pylint: disable=import-outside-toplevel
        command.extend(['--add-data', f'{Path(pyecharts.__file__).parent}{os.pathsep}pyecharts'])

    command.extend([args.main])

    print('PyInstaller command:')
    print(' ', ' '.join(command))
    if args.dry_run:
        return

    subprocess.call(command)


if __name__ == '__main__':
    main()



from typing import Literal

from ..defaults import DEFAULT_PROP, resolve_defaults
from ..element import Element

PageStickyPositions = Literal[
    'top-right',
    'top-left',
    'bottom-right',
    'bottom-left',
    'top',
    'right',
    'bottom',
    'left',
]


class PageSticky(Element):

    @resolve_defaults
    def __init__(self,
                 position: PageStickyPositions = DEFAULT_PROP | 'bottom-right',
                 x_offset: float = 0,
                 y_offset: float = 0,
                 *,
                 expand: bool = DEFAULT_PROP | False,
                 ) -> None:
        """Page sticky

        This element is based on Quasar's `QPageSticky <https://quasar.dev/layout/page-sticky>`_ component.

        :param position: position on the screen (default: "bottom-right")
        :param x_offset: horizontal offset (default: 0)
        :param y_offset: vertical offset (default: 0)
        :param expand: whether to fully expand instead of shrinking to fit the content (default: ``False``, *added in version 2.1.0*)
        """
        super().__init__('q-page-sticky')
        self._props['position'] = position
        self._props['offset'] = [x_offset, y_offset]
        self._props.set_bool('expand', expand)



from ..defaults import DEFAULT_PROP, resolve_defaults
from ..element import Element
from .mixins.color_elements import QUASAR_COLORS


class Colors(Element, component='colors.js'):

    @resolve_defaults
    def __init__(self, *,
                 primary: str = DEFAULT_PROP | '#5898d4',
                 secondary: str = DEFAULT_PROP | '#26a69a',
                 accent: str = DEFAULT_PROP | '#9c27b0',
                 dark: str = DEFAULT_PROP | '#1d1d1d',
                 dark_page: str = DEFAULT_PROP | '#121212',
                 positive: str = DEFAULT_PROP | '#21ba45',
                 negative: str = DEFAULT_PROP | '#c10015',
                 info: str = DEFAULT_PROP | '#31ccec',
                 warning: str = DEFAULT_PROP | '#f2c037',
                 **custom_colors: str) -> None:
        """Color Theming

        Sets the main colors (primary, secondary, accent, ...) used by `Quasar <https://quasar.dev/style/theme-builder>`_ on a per-page basis.

        Note: This takes precedence over the global color configuration set via ``app.colors()``.

        :param primary: Primary color (default: "#5898d4")
        :param secondary: Secondary color (default: "#26a69a")
        :param accent: Accent color (default: "#9c27b0")
        :param dark: Dark color (default: "#1d1d1d")
        :param dark_page: Dark page color (default: "#121212")
        :param positive: Positive color (default: "#21ba45")
        :param negative: Negative color (default: "#c10015")
        :param info: Info color (default: "#31ccec")
        :param warning: Warning color (default: "#f2c037")
        :param custom_colors: Custom color definitions for branding (needs ``ui.colors`` to be called before custom color is ever used, *added in version 2.2.0*)
        """
        super().__init__()
        self._props['primary'] = primary
        self._props['secondary'] = secondary
        self._props['accent'] = accent
        self._props['dark'] = dark
        self._props['dark-page'] = dark_page
        self._props['positive'] = positive
        self._props['negative'] = negative
        self._props['info'] = info
        self._props['warning'] = warning
        self._props['custom-colors'] = custom_colors
        QUASAR_COLORS.update({name.replace('_', '-') for name in custom_colors})

        self._props.add_rename('dark_page', 'dark-page')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('custom_colors', 'custom-colors')  # DEPRECATED: remove in NiceGUI 4.0



from __future__ import annotations

import asyncio
from typing import Any, ClassVar


class JavaScriptRequest:
    _instances: ClassVar[dict[str, JavaScriptRequest]] = {}

    def __init__(self, request_id: str, *, timeout: float) -> None:
        self.request_id = request_id
        self._instances[request_id] = self
        self.timeout = timeout
        self._event = asyncio.Event()
        self._result: Any = None

    @classmethod
    def resolve(cls, request_id: str, result: Any) -> None:
        """Store the result of a JavaScript request and unblock the awaiter."""
        if request_id not in cls._instances:
            return
        request = cls._instances[request_id]
        request._result = result  # pylint: disable=protected-access
        request._event.set()  # pylint: disable=protected-access

    def __await__(self) -> Any:
        try:
            yield from asyncio.wait_for(self._event.wait(), self.timeout).__await__()
        except asyncio.TimeoutError as e:
            raise TimeoutError(f'JavaScript did not respond within {self.timeout:.1f} s') from e
        else:
            return self._result
        finally:
            self._instances.pop(self.request_id)



from ..element import Element


class List(Element):

    def __init__(self) -> None:
        """List

        A list element based on Quasar's `QList <https://quasar.dev/vue-components/list-and-list-items#qlist-api>`_ component.
        It provides a container for ``ui.item`` elements.
        """
        super().__init__('q-list')



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from nicegui import Client, background_tasks, ui
from nicegui.functions.navigate import Navigate

if TYPE_CHECKING:
    from .user import User


class UserNavigate(Navigate):

    def __init__(self, user: User) -> None:
        super().__init__()
        self.user = user

    def to(self, target: Callable[..., Any] | str | ui.element, new_tab: bool = False) -> None:
        if isinstance(target, ui.element):
            # NOTE navigation to an element does not do anything in the user simulation (the whole content is always visible)
            return
        path = Client.page_routes[target] if callable(target) else target
        background_tasks.create(self.user.open(path), name=f'navigate to {path}')

    def back(self) -> None:
        current = self.user.back_history.pop()
        self.user.forward_history.append(current)
        target = self.user.back_history.pop()
        background_tasks.create(self.user.open(target, clear_forward_history=False), name=f'navigate back to {target}')

    def forward(self) -> None:
        if not self.user.forward_history:
            return
        target = self.user.forward_history[0]
        del self.user.forward_history[0]
        background_tasks.create(self.user.open(target, clear_forward_history=False),
                                name=f'navigate forward to {target}')

    def reload(self) -> None:
        target = self.user.back_history.pop()
        background_tasks.create(self.user.open(target, clear_forward_history=False),
                                name=f'navigate reload to {target}')



from typing import Literal

from ..defaults import DEFAULT_PROP, resolve_defaults
from .mixins.color_elements import TextColorElement

SpinnerTypes = Literal[
    'default',
    'audio',
    'ball',
    'bars',
    'box',
    'clock',
    'comment',
    'cube',
    'dots',
    'facebook',
    'gears',
    'grid',
    'hearts',
    'hourglass',
    'infinity',
    'ios',
    'orbit',
    'oval',
    'pie',
    'puff',
    'radio',
    'rings',
    'tail',
]


class Spinner(TextColorElement):

    @resolve_defaults
    def __init__(self,
                 type: SpinnerTypes | None = 'default', *,  # pylint: disable=redefined-builtin
                 size: str = DEFAULT_PROP | '1em',
                 color: str | None = DEFAULT_PROP | 'primary',
                 thickness: float = DEFAULT_PROP | 5.0,
                 ) -> None:
        """Spinner

        This element is based on Quasar's `QSpinner <https://quasar.dev/vue-components/spinners>`_ component.

        :param type: type of spinner (e.g. "audio", "ball", "bars", ..., default: "default")
        :param size: size of the spinner (e.g. "3em", "10px", "xl", ..., default: "1em")
        :param color: color of the spinner (either a Quasar, Tailwind, or CSS color or `None`, default: "primary")
        :param thickness: thickness of the spinner (applies to the "default" spinner only, default: 5.0)
        """
        super().__init__(tag='q-spinner' if type == 'default' else f'q-spinner-{type}', text_color=color)
        self._props['size'] = size
        self._props['thickness'] = thickness



from typing import Literal

from ..element import Element


class Column(Element, default_classes='nicegui-column'):

    def __init__(self, *,
                 wrap: bool = False,
                 align_items: Literal['start', 'end', 'center', 'baseline', 'stretch'] | None = None,
                 ) -> None:
        """Column Element

        Provides a container which arranges its child in a column.

        :param wrap: whether to wrap the content (default: `False`)
        :param align_items: alignment of the items in the column ("start", "end", "center", "baseline", or "stretch"; default: `None`)
        """
        super().__init__('div')
        if align_items:
            self._classes.append(f'items-{align_items}')

        if wrap:
            self._style['flex-wrap'] = 'wrap'



from ..defaults import DEFAULT_PROP, resolve_defaults
from .mixins.color_elements import BackgroundColorElement, TextColorElement
from .mixins.icon_element import IconElement


class Avatar(IconElement, BackgroundColorElement, TextColorElement):
    TEXT_COLOR_PROP = 'text-color'

    @resolve_defaults
    def __init__(self,
                 icon: str | None = None, *,
                 color: str | None = DEFAULT_PROP | 'primary',
                 text_color: str | None = DEFAULT_PROP | None,
                 size: str | None = DEFAULT_PROP | None,
                 font_size: str | None = DEFAULT_PROP | None,
                 square: bool = DEFAULT_PROP | False,
                 rounded: bool = DEFAULT_PROP | False,
                 ) -> None:
        """Avatar

        A avatar element wrapping Quasar's
        `QAvatar <https://quasar.dev/vue-components/avatar>`_ component.

        :param icon: name of the icon or image path with "img:" prefix (e.g. "map", "img:path/to/image.png")
        :param color: background color (either a Quasar, Tailwind, or CSS color or `None`, default: "primary")
        :param text_color: color name from the Quasar Color Palette (e.g. "primary", "teal-10")
        :param size: size in CSS units, including unit name or standard size name (xs|sm|md|lg|xl) (e.g. "16px", "2rem")
        :param font_size: size in CSS units, including unit name, of the content (icon, text) (e.g. "18px", "2rem")
        :param square: removes border-radius so borders are squared (default: False)
        :param rounded: applies a small standard border-radius for a squared shape of the component (default: False)
        """
        super().__init__(tag='q-avatar', background_color=color, text_color=text_color, icon=icon)

        self._props['square'] = square
        self._props['rounded'] = rounded
        self._props.set_optional('size', size)
        self._props.set_optional('font-size', font_size)



from .language import Language

translations: dict[Language, dict[str, str]] = {
    'de': {
        'connection_lost': 'Verbindung verloren',
        'trying_to_reconnect': 'Versuche erneut zu verbinden...',
        'message_too_long': 'Nachricht zu lang',
        'message_too_long_body': 'Die Nachricht ist zu groß für die Übertragung via WebSocket.'
    },
    'de-CH': {
        'connection_lost': 'Verbindung verloren',
        'trying_to_reconnect': 'Versuche erneut zu verbinden...',
        'message_too_long': 'Nachricht zu lang',
        'message_too_long_body': 'Die Nachricht ist zu groß für die Übertragung via WebSocket.'
    },
    'de-DE': {
        'connection_lost': 'Verbindung verloren',
        'trying_to_reconnect': 'Versuche erneut zu verbinden...',
        'message_too_long': 'Nachricht zu lang',
        'message_too_long_body': 'Die Nachricht ist zu groß für die Übertragung via WebSocket.'
    },
    'en-GB': {
        'connection_lost': 'Connection lost.',
        'trying_to_reconnect': 'Trying to reconnect...',
        'message_too_long': 'Message too long',
        'message_too_long_body': 'The message is too large for WebSocket transmission.'
    },
    'en-US': {
        'connection_lost': 'Connection lost.',
        'trying_to_reconnect': 'Trying to reconnect...',
        'message_too_long': 'Message too long',
        'message_too_long_body': 'The message is too large for WebSocket transmission.'
    },
    'es': {
        'connection_lost': 'Conexión perdida',
        'trying_to_reconnect': 'Intentando reconectar...',
        'message_too_long': 'Mensaje demasiado largo',
        'message_too_long_body': 'El mensaje excede el límite de tamaño de WebSocket.'
    },
    'fr': {
        'connection_lost': 'Connexion perdue',
        'trying_to_reconnect': 'Tentative de reconnexion...',
        'message_too_long': 'Message trop long',
        'message_too_long_body': 'Le message dépasse la limite de taille WebSocket.'
    },
    'it': {
        'connection_lost': 'Connessione persa',
        'trying_to_reconnect': 'Tentativo di riconnessione...',
        'message_too_long': 'Messaggio troppo lungo',
        'message_too_long_body': 'Il messaggio supera il limite di dimensione WebSocket.'
    },
    'ja': {
        'connection_lost': '接続が切断されました',
        'trying_to_reconnect': '再接続を試みています...',
        'message_too_long': 'メッセージが長すぎます',
        'message_too_long_body': 'メッセージがWebSocketの送信制限を超えています。'
    },
    'ko-KR': {
        'connection_lost': '연결이 끊어졌습니다',
        'trying_to_reconnect': '다시 연결하는 중...',
        'message_too_long': '메시지가 너무 깁니다',
        'message_too_long_body': '메시지가 WebSocket 크기 제한을 초과합니다.'
    },
    'ru': {
        'connection_lost': 'Соединение потеряно',
        'trying_to_reconnect': 'Попытка переподключения...',
        'message_too_long': 'Сообщение слишком длинное',
        'message_too_long_body': 'Сообщение превышает ограничение размера WebSocket.'
    },
    'zh-CN': {
        'connection_lost': '连接丢失',
        'trying_to_reconnect': '正在尝试重新连接...',
        'message_too_long': '消息太长',
        'message_too_long_body': '消息超过WebSocket大小限制。'
    },
    'zh-TW': {
        'connection_lost': '連線丟失',
        'trying_to_reconnect': '正在嘗試重新連線...',
        'message_too_long': '訊息過長',
        'message_too_long_body': '訊息超過WebSocket大小限制。'
    },
}



from __future__ import annotations

import math
import uuid
from typing import TYPE_CHECKING, Any, Literal

from typing_extensions import Self

if TYPE_CHECKING:
    from .scene import Scene, SceneObject


class Object3D:
    current_scene: Scene | None = None

    def __init__(self, type_: str, *args: Any) -> None:
        self.type = type_
        self.id = str(uuid.uuid4())
        self.name: str | None = None
        assert self.current_scene is not None
        self.scene: Scene = self.current_scene
        self.scene.objects[self.id] = self
        self.parent: Object3D | SceneObject = self.scene.stack[-1]
        self.args: list = list(args)
        self.color: str | None = '#ffffff'
        self.opacity: float = 1.0
        self.side_: str = 'front'
        self.visible_: bool = True
        self.draggable_: bool = False
        self.x: float = 0
        self.y: float = 0
        self.z: float = 0
        self.R: list[list[float]] = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        self.sx: float = 1
        self.sy: float = 1
        self.sz: float = 1
        self._create()

    def with_name(self, name: str) -> Self:
        """Set the name of the object."""
        self.name = name
        self._name()
        return self

    @property
    def data(self) -> list[Any]:
        """Data to be sent to the frontend."""
        return [
            self.type, self.id, self.parent.id, self.args,
            self.name,
            self.color, self.opacity, self.side_,
            self.x, self.y, self.z,
            self.R,
            self.sx, self.sy, self.sz,
            self.visible_,
            self.draggable_,
        ]

    def __enter__(self) -> Self:
        self.scene.stack.append(self)
        return self

    def __exit__(self, *_) -> None:
        self.scene.stack.pop()

    def _create(self) -> None:
        self.scene.run_method('create', self.type, self.id, self.parent.id, *self.args)

    def _name(self) -> None:
        self.scene.run_method('name', self.id, self.name)

    def _material(self) -> None:
        self.scene.run_method('material', self.id, self.color, self.opacity, self.side_)

    def _move(self) -> None:
        self.scene.run_method('move', self.id, self.x, self.y, self.z)

    def _rotate(self) -> None:
        self.scene.run_method('rotate', self.id, self.R)

    def _scale(self) -> None:
        self.scene.run_method('scale', self.id, self.sx, self.sy, self.sz)

    def _visible(self) -> None:
        self.scene.run_method('visible', self.id, self.visible_)

    def _draggable(self) -> None:
        self.scene.run_method('draggable', self.id, self.draggable_)

    def _delete(self) -> None:
        self.scene.run_method('delete', self.id)

    def material(self,
                 color: str | None = '#ffffff',
                 opacity: float = 1.0,
                 side: Literal['front', 'back', 'both'] = 'front',
                 ) -> Self:
        """Set the color and opacity of the object.

        :param color: CSS color string (default: '#ffffff')
        :param opacity: opacity between 0.0 and 1.0 (default: 1.0)
        :param side: 'front', 'back', or 'double' (default: 'front')
        """
        if self.color != color or self.opacity != opacity or self.side_ != side:
            self.color = color
            self.opacity = opacity
            self.side_ = side
            self._material()
        return self

    def move(self, x: float = 0.0, y: float = 0.0, z: float = 0.0) -> Self:
        """Move the object.

        :param x: x coordinate
        :param y: y coordinate
        :param z: z coordinate
        """
        if self.x != x or self.y != y or self.z != z:
            self.x = x
            self.y = y
            self.z = z
            self._move()
        return self

    @staticmethod
    def rotation_matrix_from_euler(r_x: float, r_y: float, r_z: float) -> list[list[float]]:
        """Create a rotation matrix from Euler angles.

        :param r_x: rotation around the x axis in radians
        :param r_y: rotation around the y axis in radians
        :param r_z: rotation around the z axis in radians
        """
        sx, cx = math.sin(r_x), math.cos(r_x)
        sy, cy = math.sin(r_y), math.cos(r_y)
        sz, cz = math.sin(r_z), math.cos(r_z)
        return [
            [cz * cy, -sz * cx + cz * sy * sx, sz * sx + cz * sy * cx],
            [sz * cy, cz * cx + sz * sy * sx, -cz * sx + sz * sy * cx],
            [-sy, cy * sx, cy * cx],
        ]

    def rotate(self, r_x: float, r_y: float, r_z: float) -> Self:
        """Rotate the object.

        :param r_x: rotation around the x axis in radians
        :param r_y: rotation around the y axis in radians
        :param r_z: rotation around the z axis in radians
        """
        return self.rotate_R(self.rotation_matrix_from_euler(r_x, r_y, r_z))

    def rotate_R(self, R: list[list[float]]) -> Self:
        """Rotate the object.

        :param R: 3x3 rotation matrix
        """
        if self.R != R:
            self.R = R
            self._rotate()
        return self

    def scale(self, sx: float = 1.0, sy: float | None = None, sz: float | None = None) -> Self:
        """Scale the object.

        :param sx: scale factor for the x axis
        :param sy: scale factor for the y axis (default: `sx`)
        :param sz: scale factor for the z axis (default: `sx`)
        """
        if sy is None:
            sy = sx
        if sz is None:
            sz = sx
        if self.sx != sx or self.sy != sy or self.sz != sz:
            self.sx = sx
            self.sy = sy
            self.sz = sz
            self._scale()
        return self

    def visible(self, value: bool = True) -> Self:
        """Set the visibility of the object.

        :param value: whether the object should be visible (default: `True`)
        """
        if self.visible_ != value:
            self.visible_ = value
            self._visible()
        return self

    def draggable(self, value: bool = True) -> Self:
        """Set whether the object should be draggable.

        :param value: whether the object should be draggable (default: `True`)
        """
        if self.draggable_ != value:
            self.draggable_ = value
            self._draggable()
        return self

    def attach(self, parent: Object3D) -> None:
        """Attach the object to a parent object.

        The position and rotation of the object are preserved so that the object does not move in space.

        But note that scaling is not preserved.
        If either the parent or the object itself is scaled, the object shape and position can change.

        *Added in version 2.7.0*
        """
        self.detach()
        self.parent = parent
        self._move_into_parent(parent)
        self.scene.run_method('attach', self.id, parent.id, self.x, self.y, self.z, self.R)

    def _move_into_parent(self, parent: Object3D | SceneObject) -> None:
        if not isinstance(parent, Object3D):
            return
        if isinstance(parent.parent, Object3D):
            self._move_into_parent(parent.parent)
        M1: list[list[float]] = [
            [self.R[0][0], self.R[0][1], self.R[0][2], self.x],
            [self.R[1][0], self.R[1][1], self.R[1][2], self.y],
            [self.R[2][0], self.R[2][1], self.R[2][2], self.z],
            [0, 0, 0, 1],
        ]
        M2_inv: list[list[float]] = [
            [parent.R[0][0], parent.R[1][0], parent.R[2][0],
             - parent.R[0][0] * parent.x
             - parent.R[1][0] * parent.y
             - parent.R[2][0] * parent.z],
            [parent.R[0][1], parent.R[1][1], parent.R[2][1],
             - parent.R[0][1] * parent.x
             - parent.R[1][1] * parent.y
             - parent.R[2][1] * parent.z],
            [parent.R[0][2], parent.R[1][2], parent.R[2][2],
             - parent.R[0][2] * parent.x
             - parent.R[1][2] * parent.y
             - parent.R[2][2] * parent.z],
            [0, 0, 0, 1],
        ]
        M: list[list[float]] = [
            [
                M2_inv[0][0] * M1[0][0] + M2_inv[0][1] * M1[1][0] + M2_inv[0][2] * M1[2][0],
                M2_inv[0][0] * M1[0][1] + M2_inv[0][1] * M1[1][1] + M2_inv[0][2] * M1[2][1],
                M2_inv[0][0] * M1[0][2] + M2_inv[0][1] * M1[1][2] + M2_inv[0][2] * M1[2][2],
                M2_inv[0][0] * M1[0][3] + M2_inv[0][1] * M1[1][3] + M2_inv[0][2] * M1[2][3] + M2_inv[0][3],
            ],
            [
                M2_inv[1][0] * M1[0][0] + M2_inv[1][1] * M1[1][0] + M2_inv[1][2] * M1[2][0],
                M2_inv[1][0] * M1[0][1] + M2_inv[1][1] * M1[1][1] + M2_inv[1][2] * M1[2][1],
                M2_inv[1][0] * M1[0][2] + M2_inv[1][1] * M1[1][2] + M2_inv[1][2] * M1[2][2],
                M2_inv[1][0] * M1[0][3] + M2_inv[1][1] * M1[1][3] + M2_inv[1][2] * M1[2][3] + M2_inv[1][3],
            ],
            [
                M2_inv[2][0] * M1[0][0] + M2_inv[2][1] * M1[1][0] + M2_inv[2][2] * M1[2][0],
                M2_inv[2][0] * M1[0][1] + M2_inv[2][1] * M1[1][1] + M2_inv[2][2] * M1[2][1],
                M2_inv[2][0] * M1[0][2] + M2_inv[2][1] * M1[1][2] + M2_inv[2][2] * M1[2][2],
                M2_inv[2][0] * M1[0][3] + M2_inv[2][1] * M1[1][3] + M2_inv[2][2] * M1[2][3] + M2_inv[2][3],
            ],
            [
                0, 0, 0, 1,
            ],
        ]
        self.x = M[0][3]
        self.y = M[1][3]
        self.z = M[2][3]
        self.R = [
            [M[0][0], M[0][1], M[0][2]],
            [M[1][0], M[1][1], M[1][2]],
            [M[2][0], M[2][1], M[2][2]],
        ]

    def detach(self) -> None:
        """Remove the object from its parent group object.

        The position and rotation of the object are preserved so that the object does not move in space.

        But note that scaling is not preserved.
        If either the parent or the object itself is scaled, the object shape and position can change.

        *Added in version 2.7.0*
        """
        self._move_out_of_parent(self.parent)
        self.parent = self.scene.stack[0]
        self.scene.run_method('detach', self.id, self.x, self.y, self.z, self.R)

    def _move_out_of_parent(self, parent: Object3D | SceneObject) -> None:
        if not isinstance(parent, Object3D):
            return
        M1: list[list[float]] = [
            [self.R[0][0], self.R[0][1], self.R[0][2], self.x],
            [self.R[1][0], self.R[1][1], self.R[1][2], self.y],
            [self.R[2][0], self.R[2][1], self.R[2][2], self.z],
            [0, 0, 0, 1],
        ]
        M2: list[list[float]] = [
            [parent.R[0][0], parent.R[0][1], parent.R[0][2], parent.x],
            [parent.R[1][0], parent.R[1][1], parent.R[1][2], parent.y],
            [parent.R[2][0], parent.R[2][1], parent.R[2][2], parent.z],
            [0, 0, 0, 1],
        ]
        M: list[list[float]] = [
            [
                M2[0][0] * M1[0][0] + M2[0][1] * M1[1][0] + M2[0][2] * M1[2][0],
                M2[0][0] * M1[0][1] + M2[0][1] * M1[1][1] + M2[0][2] * M1[2][1],
                M2[0][0] * M1[0][2] + M2[0][1] * M1[1][2] + M2[0][2] * M1[2][2],
                M2[0][0] * M1[0][3] + M2[0][1] * M1[1][3] + M2[0][2] * M1[2][3] + M2[0][3],
            ],
            [
                M2[1][0] * M1[0][0] + M2[1][1] * M1[1][0] + M2[1][2] * M1[2][0],
                M2[1][0] * M1[0][1] + M2[1][1] * M1[1][1] + M2[1][2] * M1[2][1],
                M2[1][0] * M1[0][2] + M2[1][1] * M1[1][2] + M2[1][2] * M1[2][2],
                M2[1][0] * M1[0][3] + M2[1][1] * M1[1][3] + M2[1][2] * M1[2][3] + M2[1][3],
            ],
            [
                M2[2][0] * M1[0][0] + M2[2][1] * M1[1][0] + M2[2][2] * M1[2][0],
                M2[2][0] * M1[0][1] + M2[2][1] * M1[1][1] + M2[2][2] * M1[2][1],
                M2[2][0] * M1[0][2] + M2[2][1] * M1[1][2] + M2[2][2] * M1[2][2],
                M2[2][0] * M1[0][3] + M2[2][1] * M1[1][3] + M2[2][2] * M1[2][3] + M2[2][3],
            ],
            [
                0, 0, 0, 1,
            ],
        ]
        self.x = M[0][3]
        self.y = M[1][3]
        self.z = M[2][3]
        self.R = [
            [M[0][0], M[0][1], M[0][2]],
            [M[1][0], M[1][1], M[1][2]],
            [M[2][0], M[2][1], M[2][2]],
        ]
        if isinstance(parent.parent, Object3D):
            self._move_out_of_parent(parent.parent)

    @property
    def children(self) -> list[Object3D]:
        """List of children of the object.

        *Added in version 2.4.0*
        """
        return [object for object in self.scene.objects.values() if object.parent == self]

    def delete(self) -> None:
        """Delete the object."""
        for child in self.children:
            child.delete()
        del self.scene.objects[self.id]
        self._delete()



from ..context import context
from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..helpers import require_top_level_layout
from .mixins.value_element import ValueElement


class Footer(ValueElement, default_classes='nicegui-footer'):

    @resolve_defaults
    def __init__(self, *,
                 value: bool = DEFAULT_PROPS['model-value'] | True,
                 fixed: bool = True,
                 bordered: bool = DEFAULT_PROP | False,
                 elevated: bool = DEFAULT_PROP | False,
                 wrap: bool = True,
                 ) -> None:
        """Footer

        This element is based on Quasar's `QFooter <https://quasar.dev/layout/header-and-footer#qfooter-api>`_ component.

        Like other layout elements, the footer can not be nested inside other elements.

        Note: The footer is automatically placed below other layout elements in the DOM to improve accessibility.
        To change the order, use the `move` method.

        :param value: whether the footer is already opened (default: `True`)
        :param fixed: whether the footer is fixed or scrolls with the content (default: `True`)
        :param bordered: whether the footer should have a border (default: `False`)
        :param elevated: whether the footer should have a shadow (default: `False`)
        :param wrap: whether the footer should wrap its content (default: `True`)
        """
        require_top_level_layout(self)
        with context.client.layout:
            super().__init__(tag='q-footer', value=value, on_value_change=None)
        self._props['bordered'] = bordered
        self._props['elevated'] = elevated
        if wrap:
            self._classes.append('wrap')
        code = list(self.client.layout.props['view'])
        code[9] = 'F' if fixed else 'f'
        self.client.layout.props['view'] = ''.join(code)

        self.move(target_index=-1)

    def toggle(self) -> None:
        """Toggle the footer"""
        self.value = not self.value

    def show(self) -> None:
        """Show the footer"""
        self.value = True

    def hide(self) -> None:
        """Hide the footer"""
        self.value = False



from collections.abc import Callable
from typing import Any
from urllib.parse import urlparse

from .. import background_tasks, json
from ..client import Client
from ..context import context
from ..element import Element
from ..elements.sub_pages import SubPages
from .javascript import run_javascript


class Navigate:
    """Navigation functions

    These functions allow you to navigate within the browser history and to external URLs.

    *Added in version 2.0.0*
    """

    def __init__(self) -> None:
        self.history = History()

    def back(self) -> None:
        """ui.navigate.back

        Navigates back in the browser history.
        It is equivalent to clicking the back button in the browser.
        """
        run_javascript('history.back()')

    def forward(self) -> None:
        """ui.navigate.forward

        Navigates forward in the browser history.
        It is equivalent to clicking the forward button in the browser.
        """
        run_javascript('history.forward()')

    def reload(self) -> None:
        """ui.navigate.reload

        Reload the current page.
        It is equivalent to clicking the reload button in the browser.
        """
        run_javascript('history.go(0)')

    def to(self, target: Callable[..., Any] | str | Element, new_tab: bool = False) -> None:
        """ui.navigate.to (formerly ui.open)

        Can be used to programmatically open a different page or URL.

        When using the `new_tab` parameter, the browser might block the new tab.
        This is a browser setting and cannot be changed by the application.
        You might want to use `ui.link` and its `new_tab` parameter instead.

        :param target: page function, NiceGUI element on the same page or string that is a an absolute URL or relative path from base URL
        :param new_tab: whether to open the target in a new tab (might be blocked by the browser)
        """
        if isinstance(target, str):
            path = target
        elif isinstance(target, Element):
            path = f'#{target.html_id}'
        elif callable(target):
            path = Client.page_routes[target]
        else:
            raise TypeError(f'Invalid target type: {type(target)}')

        if not new_tab and isinstance(target, str):
            parsed = urlparse(path)
            if not parsed.scheme and not parsed.netloc and \
                    any(isinstance(el, SubPages) for el in context.client.elements.values()):
                async def navigate_sub_pages(client: Client) -> None:
                    with client:
                        await client.sub_pages_router._handle_navigate(path)  # pylint: disable=protected-access
                background_tasks.create(navigate_sub_pages(context.client), name='navigate_sub_pages')
                return

        context.client.open(path, new_tab)


class History:

    def push(self, url: str) -> None:
        """Push a URL to the browser navigation history.

        See JavaScript's `pushState <https://developer.mozilla.org/en-US/docs/Web/API/History/pushState>`_ for more information.

        *Added in version 2.13.0*

        :param url: relative or absolute URL
        """
        run_javascript(f'history.pushState({{}}, "", {json.dumps(url)});')

    def replace(self, url: str) -> None:
        """Replace the current URL in the browser history.

        See JavaScript's `replaceState <https://developer.mozilla.org/en-US/docs/Web/API/History/replaceState>`_ for more information.

        *Added in version 2.13.0*

        :param url: relative or absolute URL
        """
        run_javascript(f'history.replaceState({{}}, "", {json.dumps(url)});')


navigate = Navigate()



try:
    from .screen import Screen
except ImportError:
    # we simply define Screen as None if selenium is not installed
    # this allows simpler dependency management when only using the "User" fixture
    # (see discussion in #3510)
    Screen = None  # type: ignore

from .user import User
from .user_interaction import UserInteraction
from .user_simulation import user_simulation

__all__ = [
    'Screen',
    'User',
    'UserInteraction',
    'user_simulation',
]



from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import TYPE_CHECKING

from socketio import AsyncServer

if TYPE_CHECKING:
    from .air import Air
    from .app import App
    from .client import Client

app: App
sio: AsyncServer
loop: asyncio.AbstractEventLoop | None = None
air: Air | None = None
root: Callable | None = None
script_mode: bool = False
script_client: Client | None = None


def is_script_mode_preflight() -> bool:
    """Return whether this is the preflight run of the script mode."""
    return script_mode and not app.is_started  # pylint: disable=undefined-variable # noqa: F821


def reset() -> None:
    """Reset core variables. (Useful for testing.)"""
    global loop, air, root, script_mode, script_client  # pylint: disable=global-statement # noqa: PLW0603
    loop = None
    air = None
    root = None
    script_mode = False
    script_client = None



import asyncio

from typing_extensions import Self

from ..defaults import DEFAULT_PROP, resolve_defaults
from ..events import ClickEventArguments, Handler, handle_event
from .mixins.color_elements import BackgroundColorElement
from .mixins.disableable_element import DisableableElement
from .mixins.icon_element import IconElement
from .mixins.text_element import TextElement


class Button(IconElement, TextElement, DisableableElement, BackgroundColorElement):

    @resolve_defaults
    def __init__(self,
                 text: str = '', *,
                 on_click: Handler[ClickEventArguments] | None = None,
                 color: str | None = DEFAULT_PROP | 'primary',
                 icon: str | None = DEFAULT_PROP | None,
                 ) -> None:
        """Button

        This element is based on Quasar's `QBtn <https://quasar.dev/vue-components/button>`_ component.

        The ``color`` parameter accepts a Quasar color, a Tailwind color, or a CSS color.
        If a Quasar color is used, the button will be styled according to the Quasar theme including the color of the text.
        Note that there are colors like "red" being both a Quasar color and a CSS color.
        In such cases the Quasar color will be used.

        :param text: the label of the button
        :param on_click: callback which is invoked when button is pressed
        :param color: the color of the button (either a Quasar, Tailwind, or CSS color or `None`, default: 'primary')
        :param icon: the name of an icon to be displayed on the button (default: `None`)
        """
        super().__init__(tag='q-btn', text=text, background_color=color, icon=icon)

        if on_click:
            self.on_click(on_click)

    def on_click(self, callback: Handler[ClickEventArguments]) -> Self:
        """Add a callback to be invoked when the button is clicked."""
        self.on('click', lambda _: handle_event(callback, ClickEventArguments(sender=self, client=self.client)), [])
        return self

    def _text_to_model_text(self, text: str) -> None:
        self._props['label'] = text

    async def clicked(self) -> None:
        """Wait until the button is clicked."""
        event = asyncio.Event()
        self.on('click', event.set, [])
        await self.client.connected()
        await event.wait()



from ..defaults import DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.text_element import TextElement
from .mixins.value_element import ValueElement


class Switch(TextElement, ValueElement, DisableableElement):

    @resolve_defaults
    def __init__(self,
                 text: str = '', *,
                 value: bool | None = DEFAULT_PROPS['model-value'] | False,
                 on_change: Handler[ValueChangeEventArguments] | None = None) -> None:
        """Switch

        This element is based on Quasar's `QToggle <https://quasar.dev/vue-components/toggle>`_ component.

        :param text: the label to display next to the switch
        :param value: whether it should be active initially (default: `False`)
        :param on_change: callback which is invoked when state is changed by the user
        """
        super().__init__(tag='q-toggle', text=text, value=value, on_value_change=on_change)



from .element import Element


def _create_html_element(tag: str):
    class HTMLElement(Element):
        def __init__(self, inner_html: str | None = None, **kwargs) -> None:
            super().__init__(tag)
            self._text = inner_html
            self.props.update(**kwargs)
    return HTMLElement


a = _create_html_element('a')
abbr = _create_html_element('abbr')
acronym = _create_html_element('acronym')
address = _create_html_element('address')
area = _create_html_element('area')
article = _create_html_element('article')
aside = _create_html_element('aside')
audio = _create_html_element('audio')
b = _create_html_element('b')
basefont = _create_html_element('basefont')
bdi = _create_html_element('bdi')
bdo = _create_html_element('bdo')
big = _create_html_element('big')
blockquote = _create_html_element('blockquote')
br = _create_html_element('br')
button = _create_html_element('button')
canvas = _create_html_element('canvas')
caption = _create_html_element('caption')
cite = _create_html_element('cite')
code = _create_html_element('code')
col = _create_html_element('col')
colgroup = _create_html_element('colgroup')
data = _create_html_element('data')
datalist = _create_html_element('datalist')
dd = _create_html_element('dd')
del_ = _create_html_element('del')
details = _create_html_element('details')
dfn = _create_html_element('dfn')
dialog = _create_html_element('dialog')
div = _create_html_element('div')
dl = _create_html_element('dl')
dt = _create_html_element('dt')
em = _create_html_element('em')
embed = _create_html_element('embed')
fieldset = _create_html_element('fieldset')
figcaption = _create_html_element('figcaption')
figure = _create_html_element('figure')
footer = _create_html_element('footer')
form = _create_html_element('form')
h1 = _create_html_element('h1')
header = _create_html_element('header')
hgroup = _create_html_element('hgroup')
hr = _create_html_element('hr')
i = _create_html_element('i')
iframe = _create_html_element('iframe')
img = _create_html_element('img')
input_ = _create_html_element('input')
ins = _create_html_element('ins')
kbd = _create_html_element('kbd')
label = _create_html_element('label')
legend = _create_html_element('legend')
li = _create_html_element('li')
main = _create_html_element('main')
map_ = _create_html_element('map')
mark = _create_html_element('mark')
menu = _create_html_element('menu')
meter = _create_html_element('meter')
nav = _create_html_element('nav')
object_ = _create_html_element('object')
ol = _create_html_element('ol')
optgroup = _create_html_element('optgroup')
option = _create_html_element('option')
output = _create_html_element('output')
p = _create_html_element('p')
param = _create_html_element('param')
picture = _create_html_element('picture')
pre = _create_html_element('pre')
progress = _create_html_element('progress')
q = _create_html_element('q')
rp = _create_html_element('rp')
rt = _create_html_element('rt')
ruby = _create_html_element('ruby')
s = _create_html_element('s')
samp = _create_html_element('samp')
search = _create_html_element('search')
section = _create_html_element('section')
select = _create_html_element('select')
small = _create_html_element('small')
source = _create_html_element('source')
span = _create_html_element('span')
strong = _create_html_element('strong')
sub = _create_html_element('sub')
summary = _create_html_element('summary')
sup = _create_html_element('sup')
svg = _create_html_element('svg')
table = _create_html_element('table')
tbody = _create_html_element('tbody')
td = _create_html_element('td')
template = _create_html_element('template')
textarea = _create_html_element('textarea')
tfoot = _create_html_element('tfoot')
th = _create_html_element('th')
thead = _create_html_element('thead')
time = _create_html_element('time')
tr = _create_html_element('tr')
track = _create_html_element('track')
u = _create_html_element('u')
ul = _create_html_element('ul')
var = _create_html_element('var')
video = _create_html_element('video')
wbr = _create_html_element('wbr')



import contextvars
import os
import uuid
from datetime import timedelta
from pathlib import Path
from typing import Any

from fastapi import FastAPI
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.middleware.sessions import SessionMiddleware
from starlette.requests import Request
from starlette.responses import Response

from . import core, helpers, observables
from .context import context
from .observables import ObservableDict
from .persistence import FilePersistentDict, PersistentDict, ReadOnlyDict, RedisPersistentDict
from .persistence.pseudo_persistent_dict import PseudoPersistentDict

request_contextvar: contextvars.ContextVar[Request | None] = contextvars.ContextVar('request_var', default=None)

GENERAL_ID = 'general'
USER_PREFIX = 'user-'
TAB_PREFIX = 'tab-'
TTL_BUFFER_SECONDS = 20  # Buffer to avoid race with prune_tab_storage polling


class RequestTrackingMiddleware(BaseHTTPMiddleware):

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        request_contextvar.set(request)
        if 'id' not in request.session:
            request.session['id'] = str(uuid.uuid4())
        request.state.responded = False
        session_id = request.session['id']
        if session_id not in core.app.storage._users:  # pylint: disable=protected-access
            await core.app.storage._create_user_storage(session_id)  # pylint: disable=protected-access
        response = await call_next(request)
        request.state.responded = True
        return response


def set_storage_secret(storage_secret: str | None = None,
                       session_middleware_kwargs: dict[str, Any] | None = None, *,
                       parent_app: FastAPI | None = None) -> None:
    """Set storage_secret and add request tracking middleware.

    If a ``parent_app`` with SessionMiddleware is provided, its secret is reused
    and no additional SessionMiddleware is added.
    A warning is issued if a different ``storage_secret`` was provided explicitly.
    """
    if parent_app and (parent_sm := next((m for m in parent_app.user_middleware if m.cls == SessionMiddleware), None)):
        parent_secret = str(parent_sm.kwargs['secret_key'] if 'secret_key' in parent_sm.kwargs else parent_sm.args[0])
        if storage_secret is not None and storage_secret != parent_secret:
            helpers.warn_once('Ignoring storage_secret because the parent app already has SessionMiddleware')
        if session_middleware_kwargs:
            helpers.warn_once('Ignoring session_middleware_kwargs because the parent app already has SessionMiddleware')
        Storage.secret = parent_secret
        core.app.user_middleware.append(Middleware(RequestTrackingMiddleware))
    else:
        if any(m.cls == SessionMiddleware for m in core.app.user_middleware):
            # NOTE not using "add_middleware" because it would be the wrong order
            core.app.user_middleware.append(Middleware(RequestTrackingMiddleware))
        elif storage_secret is not None:
            core.app.add_middleware(RequestTrackingMiddleware)
            core.app.add_middleware(SessionMiddleware, secret_key=storage_secret, **(session_middleware_kwargs or {}))
        Storage.secret = storage_secret


class Storage:
    secret: str | None = None
    '''Secret key for session storage.'''

    path = Path(os.environ.get('NICEGUI_STORAGE_PATH', '.nicegui')).resolve()
    '''Path to use for local persistence. Defaults to ".nicegui".'''

    redis_url = os.environ.get('NICEGUI_REDIS_URL', None)
    '''URL to use for shared persistent storage via Redis. Defaults to None, which means local file storage is used.'''

    redis_key_prefix = os.environ.get('NICEGUI_REDIS_KEY_PREFIX', 'nicegui:')
    '''Prefix for Redis keys. Defaults to "nicegui:".'''

    max_tab_storage_age: float = timedelta(days=30).total_seconds()
    '''Maximum age in seconds before tab storage is automatically purged. Defaults to 30 days.'''

    def __init__(self) -> None:
        self._general = Storage._create_persistent_dict(GENERAL_ID)
        self._users: dict[str, PersistentDict] = {}
        self._tabs: dict[str, ObservableDict] = {}

    @staticmethod
    def _create_persistent_dict(id: str) -> PersistentDict:  # pylint: disable=redefined-builtin
        if Storage.redis_url:
            ttl = int(core.app.storage.max_tab_storage_age + TTL_BUFFER_SECONDS) if id.startswith(TAB_PREFIX) else None
            return RedisPersistentDict(url=Storage.redis_url, id=id, key_prefix=Storage.redis_key_prefix, ttl=ttl)
        else:
            return FilePersistentDict(Storage.path / f'storage-{id}.json', encoding='utf-8')

    @property
    def browser(self) -> ReadOnlyDict | dict:
        """Small storage that is saved directly within the user's browser (encrypted cookie).

        The data is shared between all browser tabs and can only be modified before the initial request has been submitted.
        Therefore it is normally better to use `app.storage.user` instead,
        which can be modified anytime, reduces overall payload, improves security and has larger storage capacity.
        """
        if core.is_script_mode_preflight():
            return {}
        request: Request | None = request_contextvar.get()
        if request is None:
            if Storage.secret is None:
                raise RuntimeError('app.storage.browser needs a storage_secret passed in ui.run()')
            raise RuntimeError('app.storage.browser can only be used within a UI context')
        if request.state.responded:
            return ReadOnlyDict(
                request.session,
                'the response to the browser has already been built, so modifications cannot be sent back anymore'
            )
        return request.session

    @property
    def user(self) -> PersistentDict:
        """Individual user storage that is persisted on the server (where NiceGUI is executed).

        The data is stored on the server.
        It is shared between all browser tabs by identifying the user via session cookie ID.
        """
        if core.is_script_mode_preflight():
            return PseudoPersistentDict()
        request: Request | None = request_contextvar.get()
        if request is None:
            if Storage.secret is None:
                raise RuntimeError('app.storage.user needs a storage_secret passed in ui.run()')
            raise RuntimeError('app.storage.user can only be used within a UI context')
        session_id = request.session['id']
        assert session_id in self._users, f'user storage for {session_id} should be created before accessing it'
        return self._users[session_id]

    async def _create_user_storage(self, session_id: str) -> None:
        self._users[session_id] = Storage._create_persistent_dict(f'{USER_PREFIX}{session_id}')
        await self._users[session_id].initialize()

    @property
    def general(self) -> PersistentDict:
        """General storage shared between all users that is persisted on the server (where NiceGUI is executed)."""
        return self._general

    @property
    def client(self) -> ObservableDict:
        """A volatile storage that is only kept during the current connection to the client.

        Like `app.storage.tab` data is unique per browser tab but is even more volatile as it is already discarded
        when the connection to the client is lost through a page reload or a navigation.
        """
        return context.client.storage

    @property
    def tab(self) -> observables.ObservableDict:
        """A volatile storage that is only kept during the current tab session."""
        client = context.client
        if not client.has_socket_connection:
            raise RuntimeError('app.storage.tab can only be used with a client connection; '
                               'see https://nicegui.io/documentation/page#wait_for_client_connection to await it')
        assert client.tab_id is not None
        assert client.tab_id in self._tabs, f'tab storage for {client.tab_id} should be created before accessing it'
        return self._tabs[client.tab_id]

    async def _create_tab_storage(self, tab_id: str) -> None:
        """Create tab storage for the given tab ID."""
        if tab_id not in self._tabs:
            if Storage.redis_url:
                self._tabs[tab_id] = Storage._create_persistent_dict(f'{TAB_PREFIX}{tab_id}')
                tab = self._tabs[tab_id]
                assert isinstance(tab, PersistentDict)
                await tab.initialize()
            else:
                self._tabs[tab_id] = ObservableDict()

    def copy_tab(self, old_tab_id: str, tab_id: str) -> None:
        """Copy the tab storage to a new tab. (For internal use only.)"""
        if old_tab_id in self._tabs:
            if Storage.redis_url:
                self._tabs[tab_id] = Storage._create_persistent_dict(f'{TAB_PREFIX}{tab_id}')
            else:
                self._tabs[tab_id] = ObservableDict()
            self._tabs[tab_id].update(self._tabs[old_tab_id])

    async def close_tab(self, tab_id: str | None) -> None:
        """Close the tab storage. (For internal use only.)"""
        if tab_id and isinstance(tab := self._tabs.get(tab_id), PersistentDict):
            await tab.close()

    def clear(self) -> None:
        """Clears all storage."""
        self._general.clear()
        self._users.clear()
        if not helpers.is_pytest():
            context.client.storage.clear()
        self._tabs.clear()
        for filepath in self.path.glob('storage-*.json'):
            filepath.unlink()
        if self.path.exists():
            self.path.rmdir()

    async def on_shutdown(self) -> None:
        """Close all persistent storage. (For internal use only.)"""
        for user in self._users.values():
            await user.close()
        await self._general.close()



from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .button import Button as button
from .date import Date as date
from .menu import Menu as menu
from .mixins.disableable_element import DisableableElement
from .mixins.label_element import LabelElement
from .mixins.value_element import ValueElement


class DateInput(LabelElement, ValueElement, DisableableElement):
    LOOPBACK = False

    @resolve_defaults
    def __init__(self,
                 label: str | None = DEFAULT_PROP | None, *,
                 range_input: bool = False,
                 placeholder: str | None = DEFAULT_PROP | None,
                 value: str = DEFAULT_PROPS['model-value'] | '',
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Date Input

        This element extends Quasar's `QInput <https://quasar.dev/vue-components/input>`_ component with a date picker.

        *Added in version 3.3.0*

        :param label: displayed label for the date input
        :param range_input: if True, allows selecting a range of dates (value will be a dictionary with "from" and "to" keys)
        :param placeholder: text to show if no date is selected
        :param value: the current date value
        :param on_change: callback to execute when the value changes
        """
        self._range_input = range_input
        super().__init__(tag='q-input', label=label, value=value, on_value_change=on_change)
        self._props['for'] = self.html_id
        self._props.set_optional('placeholder', placeholder)

        with self.add_slot('append'):
            with button(icon='edit_calendar', color=None).props('flat round').classes('cursor-pointer') as self.button:
                with menu() as self.menu:
                    self.picker = date().props('no-parent-event').props('range' if range_input else '')

        self.picker.bind_value(self,
                               forward=lambda v: self._picker_to_input_value(v) if self._range_input else v,
                               backward=lambda v: self._input_to_picker_value(v) if self._range_input else v)

    def _picker_to_input_value(self, value: dict | str) -> str | None:
        if isinstance(value, dict):
            return f'{value["from"]} - {value["to"]}' if 'from' in value and 'to' in value else None
        else:
            return value

    def _input_to_picker_value(self, value: str | None) -> dict | str | None:
        if value and ' - ' in value:
            from_date, to_date = value.split(' - ', 1)
            return {'from': from_date, 'to': to_date}
        else:
            return value



from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.icon_element import IconElement
from .mixins.text_element import TextElement
from .mixins.value_element import ValueElement


class Expansion(IconElement, TextElement, ValueElement, DisableableElement,
                component='expansion.js', default_classes='nicegui-expansion'):

    @resolve_defaults
    def __init__(self,
                 text: str = DEFAULT_PROPS['label'] | '', *,
                 caption: str | None = DEFAULT_PROP | None,
                 icon: str | None = DEFAULT_PROP | None,
                 group: str | None = DEFAULT_PROP | None,
                 value: bool = DEFAULT_PROPS['model-value'] | False,
                 on_value_change: Handler[ValueChangeEventArguments] | None = None
                 ) -> None:
        """Expansion Element

        Provides an expandable container based on Quasar's `QExpansionItem <https://quasar.dev/vue-components/expansion-item>`_ component.

        :param text: title text
        :param caption: optional caption (or sub-label) text
        :param icon: optional icon (default: None)
        :param group: optional group name for coordinated open/close state within the group a.k.a. "accordion mode"
        :param value: whether the expansion should be opened on creation (default: `False`)
        :param on_value_change: callback to execute when value changes
        """
        super().__init__(icon=icon, text=text, value=value, on_value_change=on_value_change)
        self._props.set_optional('caption', caption)
        self._props.set_optional('group', group)

    def open(self) -> None:
        """Open the expansion."""
        self.value = True

    def close(self) -> None:
        """Close the expansion."""
        self.value = False

    def _text_to_model_text(self, text: str) -> None:
        self._props['label'] = text



from typing import Literal

from ..context import context
from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..helpers import require_top_level_layout
from .mixins.value_element import ValueElement

DrawerSides = Literal['left', 'right']


class Drawer(ValueElement, default_classes='nicegui-drawer'):

    @resolve_defaults
    def __init__(self,
                 side: DrawerSides, *,
                 value: bool | None = DEFAULT_PROPS['model-value'] | None,
                 fixed: bool = True,
                 bordered: bool = DEFAULT_PROP | False,
                 elevated: bool = DEFAULT_PROP | False,
                 top_corner: bool = False,
                 bottom_corner: bool = False) -> None:
        """Drawer

        This element is based on Quasar's `QDrawer <https://quasar.dev/layout/drawer>`_ component.

        Like other layout elements, a drawer can not be nested inside other elements.

        Note: Depending on the side, the drawer is automatically placed above or below the main page container in the DOM to improve accessibility.
        To change the order, use the `move` method.

        A value of ``None`` will automatically open or close the drawer depending on the current layout width (breakpoint: >=1024 px).
        The value will be requested from the client when the websocket connection is established.

        :param side: side of the page where the drawer should be placed (`left` or `right`)
        :param value: whether the drawer is already opened (default: `None`, i.e. if layout width is above threshold)
        :param fixed: whether the drawer is fixed or scrolls with the content (default: `True`)
        :param bordered: whether the drawer should have a border (default: `False`)
        :param elevated: whether the drawer should have a shadow (default: `False`)
        :param top_corner: whether the drawer expands into the top corner (default: `False`)
        :param bottom_corner: whether the drawer expands into the bottom corner (default: `False`)
        """
        require_top_level_layout(self)
        with context.client.layout:
            super().__init__(tag='q-drawer', value=value, on_value_change=None)
        self._props['show-if-above'] = value is None
        self._props['side'] = side
        self._props['bordered'] = bordered
        self._props['elevated'] = elevated
        code = list(self.client.layout.props['view'])
        code[0 if side == 'left' else 2] = side[0].lower() if top_corner else 'h'
        code[4 if side == 'left' else 6] = side[0].upper() if fixed else side[0].lower()
        code[8 if side == 'left' else 10] = side[0].lower() if bottom_corner else 'f'
        self.client.layout.props['view'] = ''.join(code)

        page_container_index = self.client.layout.default_slot.children.index(self.client.page_container)
        self.move(target_index=page_container_index if side == 'left' else page_container_index + 1)

        if value is None:
            async def _request_value() -> None:
                self.value = await context.client.run_javascript(
                    f'!getHtmlElement({self.id}).parentElement.classList.contains("q-layout--prevent-focus")  // __IS_DRAWER_OPEN__'
                )
            self.client.on_connect(_request_value)

    def toggle(self) -> None:
        """Toggle the drawer"""
        if self.value is None:
            self.run_method('toggle')
        else:
            self.value = not self.value

    def show(self) -> None:
        """Show the drawer"""
        self.value = True

    def hide(self) -> None:
        """Hide the drawer"""
        self.value = False

    def _handle_value_change(self, value: bool) -> None:
        super()._handle_value_change(value)
        self._props['show-if-above'] = value is None


class LeftDrawer(Drawer):

    @resolve_defaults
    def __init__(self, *,
                 value: bool | None = DEFAULT_PROPS['model-value'] | None,
                 fixed: bool = True,
                 bordered: bool = DEFAULT_PROP | False,
                 elevated: bool = DEFAULT_PROP | False,
                 top_corner: bool = False,
                 bottom_corner: bool = False) -> None:
        """Left drawer

        This element is based on Quasar's `QDrawer <https://quasar.dev/layout/drawer>`_ component.

        Like other layout elements, the left drawer can not be nested inside other elements.

        Note: The left drawer is automatically placed above the main page container in the DOM to improve accessibility.
        To change the order, use the `move` method.

        A value of ``None`` will automatically open or close the drawer depending on the current layout width (breakpoint: >=1024 px).
        The value will be requested from the client when the websocket connection is established.

        :param value: whether the drawer is already opened (default: `None`, i.e. if layout width is above threshold)
        :param fixed: whether the drawer is fixed or scrolls with the content (default: `True`)
        :param bordered: whether the drawer should have a border (default: `False`)
        :param elevated: whether the drawer should have a shadow (default: `False`)
        :param top_corner: whether the drawer expands into the top corner (default: `False`)
        :param bottom_corner: whether the drawer expands into the bottom corner (default: `False`)
        """
        super().__init__('left',
                         value=value,
                         fixed=fixed,
                         bordered=bordered,
                         elevated=elevated,
                         top_corner=top_corner,
                         bottom_corner=bottom_corner)


class RightDrawer(Drawer):

    @resolve_defaults
    def __init__(self, *,
                 value: bool | None = DEFAULT_PROPS['model-value'] | None,
                 fixed: bool = True,
                 bordered: bool = DEFAULT_PROP | False,
                 elevated: bool = DEFAULT_PROP | False,
                 top_corner: bool = False,
                 bottom_corner: bool = False) -> None:
        """Right drawer

        This element is based on Quasar's `QDrawer <https://quasar.dev/layout/drawer>`_ component.

        Like other layout elements, the right drawer can not be nested inside other elements.

        Note: The right drawer is automatically placed below the main page container in the DOM to improve accessibility.
        To change the order, use the `move` method.

        A value of ``None`` will automatically open or close the drawer depending on the current layout width (breakpoint: >=1024 px).
        The value will be requested from the client when the websocket connection is established.

        :param value: whether the drawer is already opened (default: `None`, i.e. if layout width is above threshold)
        :param fixed: whether the drawer is fixed or scrolls with the content (default: `True`)
        :param bordered: whether the drawer should have a border (default: `False`)
        :param elevated: whether the drawer should have a shadow (default: `False`)
        :param top_corner: whether the drawer expands into the top corner (default: `False`)
        :param bottom_corner: whether the drawer expands into the bottom corner (default: `False`)
        """
        super().__init__('right',
                         value=value,
                         fixed=fixed,
                         bordered=bordered,
                         elevated=elevated,
                         top_corner=top_corner,
                         bottom_corner=bottom_corner)



import re
import runpy
import threading
import time
from collections.abc import Callable, Generator
from contextlib import contextmanager, suppress
from pathlib import Path
from typing import Any, overload
from urllib.parse import urlparse

import pytest
from selenium import webdriver
from selenium.common.exceptions import (
    ElementNotInteractableException,
    NoSuchElementException,
    StaleElementReferenceException,
)
from selenium.webdriver import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement

from nicegui import app, core, ui
from nicegui.server import Server

from .general import prepare_simulation
from .general_fixtures import get_path_to_main_file


class Screen:
    PORT = 3392
    IMPLICIT_WAIT = 4
    SCREENSHOT_DIR = Path('screenshots')
    CATCH_JS_ERRORS = True

    def __init__(self, selenium: webdriver.Chrome, caplog: pytest.LogCaptureFixture, request: pytest.FixtureRequest | None = None) -> None:
        self.selenium = selenium
        self.caplog = caplog
        self.server_thread: threading.Thread | None = None
        self.pytest_request = request
        self.ui_run_kwargs: dict[str, Any] = {'port': self.PORT, 'show': False, 'reload': False}
        self.connected = threading.Event()
        app.on_connect(self.connected.set)
        self.url = f'http://localhost:{self.PORT}'
        self.allowed_js_errors: list[str] = []

    def start_server(self) -> None:
        """Start the webserver in a separate thread."""
        main_path = get_path_to_main_file(self.pytest_request) if self.pytest_request else None
        if main_path is None:
            prepare_simulation()
            self.server_thread = threading.Thread(target=lambda: ui.run(**self.ui_run_kwargs))
        else:
            self.server_thread = threading.Thread(target=lambda: runpy.run_path(str(main_path), run_name='__main__'))
        self.server_thread.start()

    @property
    def is_open(self) -> bool:
        """Check if the browser is open."""
        # https://stackoverflow.com/a/66150779/3419103
        try:
            self.selenium.current_url  # pylint: disable=pointless-statement # noqa: B018
            return True
        except Exception as e:
            print(e)
            return False

    def stop_server(self) -> None:
        """Stop the webserver."""
        self.close()
        self.caplog.clear()
        if hasattr(Server, 'instance'):
            Server.instance.should_exit = True
        if self.server_thread:
            self.server_thread.join()
        if core.loop:
            assert core.loop.is_closed()

    def open(self, path: str, timeout: float = 3.0) -> None:
        """Try to open the page until the server is ready or we time out.

        If the server is not yet running, start it.
        """
        if self.server_thread is None:
            self.start_server()
        deadline = time.time() + timeout
        self.connected.clear()
        while True:
            try:
                self.selenium.get(self.url + path)
                self.selenium.find_element(By.XPATH, '//body')  # ensure page and JS are loaded
                self.connected.wait(1)  # Ensure that the client has connected to the API
                break
            except Exception as e:
                if time.time() > deadline:
                    raise
                time.sleep(0.1)
                assert self.server_thread is not None
                if not self.server_thread.is_alive():
                    raise RuntimeError('The NiceGUI server has stopped running') from e
        self._wait_for_socket_idle()

    def close(self) -> None:
        """Close the browser tab.

        When the driver is session-scoped, closing the last window would invalidate the session.
        Instead, we navigate to about:blank to trigger disconnection.
        """
        if self.is_open:
            if len(self.selenium.window_handles) > 1:
                self.selenium.close()
            else:
                self.selenium.get('about:blank')

    def switch_to(self, tab_id: int) -> None:
        """Switch to the tab with the given index, or create it if it does not exist."""
        window_count = len(self.selenium.window_handles)
        if tab_id > window_count:
            raise IndexError(f'Could not go to or create tab {tab_id}, there are only {window_count} tabs')
        if tab_id == window_count:
            self.selenium.switch_to.new_window('tab')
        else:
            self.selenium.switch_to.window(self.selenium.window_handles[tab_id])

    def should_contain(self, text: str) -> None:
        """Assert that the page contains the given text."""
        if self.selenium.title == text:
            return
        self.find(text)

    @overload
    def wait_for(self, target: str) -> None:
        """Wait until the page contains the given text."""

    @overload
    def wait_for(self, target: Callable[..., bool]) -> None:
        """Wait until the given condition is met."""

    def wait_for(self, target: str | Callable[..., bool]) -> None:
        """Wait until the page contains the given text or the given condition is met."""
        if isinstance(target, str):
            self.should_contain(target)
        if callable(target):
            deadline = time.time() + self.IMPLICIT_WAIT
            while time.time() < deadline:
                try:
                    if target():
                        return
                except StaleElementReferenceException:
                    pass  # element became stale, retry
                self.wait(0.1)
            raise AssertionError('Condition not met')

    def should_not_contain(self, text: str, wait: float = 0.5) -> None:
        """Assert that the page does not contain the given text."""
        assert self.selenium.title != text
        try:
            self.selenium.implicitly_wait(wait)
            with pytest.raises(AssertionError):
                self.find(text)
        finally:
            self.selenium.implicitly_wait(self.IMPLICIT_WAIT)

    def should_contain_input(self, text: str) -> None:
        """Assert that the page contains an input with the given value."""
        deadline = time.time() + self.IMPLICIT_WAIT
        while time.time() < deadline:
            for input_element in self.find_all_by_tag('input'):
                if input_element.get_attribute('value') == text:
                    return
            self.wait(0.1)
        raise AssertionError(f'Could not find input with value "{text}"')

    def should_load_image(self, image: WebElement, *, timeout: float = 2.0) -> None:
        """Assert that the given image has loaded."""
        deadline = time.time() + timeout
        while time.time() < deadline:
            js = 'return arguments[0].naturalWidth > 0 && arguments[0].naturalHeight > 0'
            if self.selenium.execute_script(js, image):
                return
        raise AssertionError(f'Image not loaded: {image.get_attribute("outerHTML")}')

    def click(self, target_text: str) -> WebElement:
        """Click on the element containing the given text."""
        element = self.find(target_text)
        try:
            element.click()
        except ElementNotInteractableException as e:
            raise AssertionError(f'Could not click on "{target_text}" on:\n{element.get_attribute("outerHTML")}') from e
        return element

    def context_click(self, target_text: str) -> WebElement:
        """Right-click on the element containing the given text."""
        element = self.find(target_text)
        action = ActionChains(self.selenium)
        action.context_click(element).perform()
        return element

    def click_at_position(self, element: WebElement, x: int, y: int) -> None:
        """Click on the given element at the given position."""
        action = ActionChains(self.selenium)
        action.move_to_element_with_offset(element, x, y).click().perform()

    def type(self, text: str) -> None:
        """Type the given text into the currently focused element."""
        self.selenium.execute_script('window.focus();')
        self.wait(0.2)
        self.selenium.switch_to.active_element.send_keys(text)

    def find(self, text: str) -> WebElement:
        """Find the element containing the given text."""
        try:
            query = f'//*[not(self::script) and not(self::style) and text()[contains(., "{text}")]]'
            # HACK: repeat check after a short delay to avoid timing issue on fast machines
            for _ in range(5):
                element = self.selenium.find_element(By.XPATH, query)
                with suppress(StaleElementReferenceException):
                    if element.is_displayed():
                        return element
                self.wait(0.2)
            raise AssertionError(f'Found "{text}" but it is hidden')
        except NoSuchElementException as e:
            raise AssertionError(f'Could not find "{text}"') from e

    def find_all(self, text: str) -> list[WebElement]:
        """Find all elements containing the given text."""
        query = f'//*[not(self::script) and not(self::style) and text()[contains(., "{text}")]]'
        return self.selenium.find_elements(By.XPATH, query)

    def find_element(self, element: ui.element) -> WebElement:
        """Find the given NiceGUI element."""
        return self.selenium.find_element(By.ID, element.html_id)

    def find_by_class(self, name: str) -> WebElement:
        """Find the element with the given CSS class."""
        return self.selenium.find_element(By.CLASS_NAME, name)

    def find_all_by_class(self, name: str) -> list[WebElement]:
        """Find all elements with the given CSS class."""
        return self.selenium.find_elements(By.CLASS_NAME, name)

    def find_by_tag(self, name: str) -> WebElement:
        """Find the element with the given HTML tag."""
        return self.selenium.find_element(By.TAG_NAME, name)

    def find_all_by_tag(self, name: str) -> list[WebElement]:
        """Find all elements with the given HTML tag."""
        return self.selenium.find_elements(By.TAG_NAME, name)

    def find_by_css(self, selector: str) -> WebElement:
        """Find the element with the given CSS selector."""
        return self.selenium.find_element(By.CSS_SELECTOR, selector)

    def render_js_logs(self) -> str:
        """Render the browser console logs as a string."""
        console = '\n'.join(log['message'] for log in self.selenium.get_log('browser'))
        return f'-- console logs ---\n{console}\n---------------------'

    def wait(self, t: float) -> None:
        """Wait for the given number of seconds."""
        time.sleep(t)

    def _wait_for_socket_idle(self) -> None:
        """Wait until no new Socket.IO messages have been received for 100ms."""
        self.selenium.execute_async_script(f'''
            const callback = arguments[arguments.length - 1];
            const deadline = Date.now() + {Screen.IMPLICIT_WAIT * 1000};
            let lastId = window.nextMessageId || 0;
            let stableSince = Date.now();
            const check = () => {{
                const currentId = window.nextMessageId || 0;
                if (currentId !== lastId) {{
                    lastId = currentId;
                    stableSince = Date.now();
                }}
                if (Date.now() - stableSince >= 100 || Date.now() >= deadline) {{
                    callback();
                }} else {{
                    setTimeout(check, 20);
                }}
            }};
            setTimeout(check, 20);
        ''')

    def shot(self, name: str, *, failed: bool) -> None:
        """Take a screenshot and store it in the screenshots directory."""
        self.SCREENSHOT_DIR.mkdir(parents=True, exist_ok=True)
        if failed:
            name = f'{name}.failed'
        filename = self.SCREENSHOT_DIR / f'{name}.png'
        print(f'Storing screenshot to {filename}')
        self.selenium.get_screenshot_as_file(filename)

    def assert_py_logger(self, level: str, message: str | re.Pattern) -> None:
        """Assert that the Python logger has received a message with the given level and text or regex pattern."""
        try:
            assert self.caplog.records, 'Expected a log message'
            record = self.caplog.records[0]
            print(record.levelname, record.message)
            assert record.levelname.strip() == level, f'Expected "{level}" but got "{record.levelname}"'

            if isinstance(message, re.Pattern):
                assert message.search(record.message), f'Expected regex "{message}" but got "{record.message}"'
            else:
                assert record.message.strip() == message, f'Expected "{message}" but got "{record.message}"'
        finally:
            self.caplog.records.pop(0)

    @contextmanager
    def implicitly_wait(self, t: float) -> Generator[None, None, None]:
        """Temporarily change the implicit wait time."""
        self.selenium.implicitly_wait(t)
        yield
        self.selenium.implicitly_wait(self.IMPLICIT_WAIT)

    @property
    def current_path(self) -> str:
        """The current path of the browser."""
        parsed = urlparse(self.selenium.current_url)
        result = parsed.path
        if parsed.query:
            result += '?' + parsed.query
        if parsed.fragment:
            result += '#' + parsed.fragment
        return result



from typing import Any, Literal

from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import GenericEventArguments, Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.label_element import LabelElement
from .mixins.validation_element import ValidationDict, ValidationElement, ValidationFunction


class InputChips(LabelElement, ValidationElement, DisableableElement):

    @resolve_defaults
    def __init__(self,
                 label: str | None = DEFAULT_PROP | None,
                 *,
                 value: list[str] | None = DEFAULT_PROPS['model-value'] | None,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 new_value_mode: Literal['add', 'add-unique', 'toggle'] = DEFAULT_PROP | 'toggle',
                 clearable: bool = DEFAULT_PROP | False,
                 validation: ValidationFunction | ValidationDict | None = None,
                 ) -> None:
        """Input Chips

        An input field that manages a collection of values as visual "chips" or tags.
        Users can type to add new chips and remove existing ones by clicking or using keyboard shortcuts.

        This element is based on Quasar's `QSelect <https://quasar.dev/vue-components/select>`_ component.
        Unlike a traditional dropdown selection, this variant focuses on free-form text input with chips,
        making it ideal for tags, keywords, or any list of user-defined values.

        You can use the ``validation`` parameter to define a dictionary of validation rules,
        e.g. ``{'Too long!': lambda value: len(value) < 3}``.
        The key of the first rule that fails will be displayed as an error message.
        Alternatively, you can pass a callable that returns an optional error message.
        To disable the automatic validation on every value change, you can use the `without_auto_validation` method.

        *Added in version 2.22.0*

        :param label: the label to display above the selection
        :param value: the initial value
        :param on_change: callback to execute when selection changes
        :param new_value_mode: handle new values from user input (default: "toggle")
        :param clearable: whether to add a button to clear the selection
        :param validation: dictionary of validation rules or a callable that returns an optional error message (default: None for no validation)
        """
        super().__init__(tag='q-select', label=label, value=value or [], on_value_change=on_change, validation=validation)

        self._props['for'] = self.html_id
        self._props['new-value-mode'] = new_value_mode
        self._props['use-input'] = True
        self._props['use-chips'] = True
        self._props['fill-input'] = True
        self._props['input-debounce'] = 0
        self._props['multiple'] = True
        self._props['hide-dropdown-icon'] = True
        self._props['clearable'] = clearable

    def _event_args_to_value(self, e: GenericEventArguments) -> Any:
        return e.args or []



from ..element import Element


class Teleport(Element, component='teleport.js'):

    def __init__(self, to: str | Element) -> None:
        """Teleport

        An element that allows us to transmit the content from within a component to any location on the page.

        :param to: NiceGUI element or CSS selector of the target element for the teleported content
        """
        super().__init__()
        if isinstance(to, Element):
            to = f'#{to.html_id}'
        self._props['to'] = to
        self._update_method = 'update'



import hashlib
import os
from collections.abc import Callable
from functools import lru_cache

import markdown2
from fastapi.responses import PlainTextResponse
from pygments.formatters import HtmlFormatter  # pylint: disable=no-name-in-module

from .. import core
from .mixins.content_element import ContentElement


class Markdown(ContentElement, component='markdown.js', default_classes='nicegui-markdown'):
    # NOTE: The Mermaid ESM is already registered in mermaid.py.

    def __init__(self,
                 content: str = '', *,
                 extras: list[str] = ['fenced-code-blocks', 'tables'],  # noqa: B006
                 sanitize: Callable[[str], str] | bool = True,
                 ) -> None:
        """Markdown Element

        Renders Markdown onto the page.

        :param content: the Markdown content to be displayed
        :param extras: list of `markdown2 extensions <https://github.com/trentm/python-markdown2/wiki/Extras#implemented-extras>`_ (default: `['fenced-code-blocks', 'tables']`)
        :param sanitize: sanitization mode:
            ``True`` (default) uses client-side sanitization via DOMPurify,
            ``False`` disables sanitization (use only with trusted content),
            or pass a callable to apply server-side sanitization
        """
        self._sanitize = sanitize
        self.extras = extras[:]
        super().__init__(content=content)
        self._props['sanitize'] = sanitize is True
        if 'mermaid' in extras:
            self._props['use-mermaid'] = True

        codehilite = self._generate_codehilite_css()
        self._props['resource-name'] = f'codehilite_{hashlib.sha256(codehilite.encode()).hexdigest()[:32]}.css'
        self.add_dynamic_resource(
            self._props['resource-name'],
            lambda: PlainTextResponse(
                codehilite,
                media_type='text/css',
                headers={'Cache-Control': core.app.config.cache_control_directives},
            ),
        )

        self._props.add_rename('resource_name', 'resource-name')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('use_mermaid', 'use-mermaid')  # DEPRECATED: remove in NiceGUI 4.0

    @staticmethod
    @lru_cache(maxsize=1)
    def _generate_codehilite_css() -> str:
        return (
            HtmlFormatter(nobackground=True).get_style_defs('.codehilite') +
            HtmlFormatter(nobackground=True, style='github-dark').get_style_defs('.body--dark .codehilite')
        )

    def _handle_content_change(self, content: str) -> None:
        html = prepare_content(content, extras=' '.join(self.extras))
        if callable(self._sanitize):
            html = self._sanitize(html)
        if self._props.get('innerHTML') != html:
            self._props['innerHTML'] = html


@lru_cache(maxsize=int(os.environ.get('MARKDOWN_CONTENT_CACHE_SIZE', '1000')))
def prepare_content(content: str, extras: str) -> str:
    """Render Markdown content to HTML."""
    return markdown2.markdown(remove_indentation(content), extras=extras.split())


def remove_indentation(text: str) -> str:
    """Remove indentation from a multi-line string based on the indentation of the first non-empty line."""
    lines = text.splitlines()
    while lines and not lines[0].strip():
        lines.pop(0)
    if not lines:
        return ''
    indentation = len(lines[0]) - len(lines[0].lstrip())
    return '\n'.join(line[indentation:] for line in lines)



from pathlib import Path

from .. import helpers, json
from ..client import Client
from ..context import context
from ..slot import Slot


def add_css(content: str | Path, *, shared: bool = False) -> None:
    """Add CSS style definitions to the page.

    This function can be used to add CSS style definitions to the head of the HTML page.

    *Added in version 2.0.0*

    :param content: CSS content (string or file path)
    :param shared: whether to add the code to all pages (default: ``False``, *added in version 2.14.0*)
    """
    if helpers.is_file(content):
        content = Path(content).read_text(encoding='utf-8')
    safe_content = json.dumps(content).replace('<', r'\u003c')
    _add_javascript(f'addStyle({safe_content});', shared=shared)


def add_scss(content: str | Path, *, indented: bool = False, shared: bool = False) -> None:  # DEPRECATED
    """Add SCSS style definitions to the page (deprecated).

    This function can be used to add SCSS style definitions to the head of the HTML page.

    *Added in version 2.0.0*

    **Note: This function is deprecated and will be removed in NiceGUI 4.0. Use add_css instead.**

    :param content: SCSS content (string or file path)
    :param indented: whether the content is indented (SASS) or not (SCSS) (default: `False`)
    :param shared: whether to add the code to all pages (default: ``False``, *added in version 2.14.0*)
    """
    content = Path(content).read_text(encoding='utf-8') if helpers.is_file(content) else str(content).strip()
    syntax = 'indented' if indented else 'scss'
    safe_content = json.dumps(content).replace('<', r'\u003c')
    _add_javascript(f'''
        import("sass").then(sass => addStyle(sass.compileString({safe_content}, {{syntax: "{syntax}"}}).css));
    ''', shared=shared)


def add_sass(content: str | Path, *, shared: bool = False) -> None:  # DEPRECATED
    """Add SASS style definitions to the page (deprecated).

    This function can be used to add SASS style definitions to the head of the HTML page.

    *Added in version 2.0.0*

    **Note: This function is deprecated and will be removed in NiceGUI 4.0. Use add_css instead.**

    :param content: SASS content (string or file path)
    :param shared: whether to add the code to all pages (default: ``False``, *added in version 2.14.0*)
    """
    add_scss(content, indented=True, shared=shared)


def _add_javascript(code: str, *, shared: bool = False) -> None:
    script_html = f'<script>{code}</script>'
    if shared:
        client = context.client if Slot.get_stack() else None  # NOTE: don't auto-create a client if shared=True
        Client.shared_head_html += script_html + '\n'
    else:
        client = context.client
        client._head_html += script_html + '\n'
    if client is not None and client.has_socket_connection:  # NOTE: no need to run JavaScript if there is no client yet
        client.run_javascript(code)



from .. import core, json
from ..context import context


def page_title(title: str) -> None:
    """Page title

    Set the page title for the current client.

    :param title: page title
    """
    client = context.client
    client.title = title
    if core.app.native.main_window:
        core.app.native.main_window.set_title(title)
    if client.has_socket_connection:
        client.run_javascript(f'document.title = {json.dumps(title)}')



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable, Iterator
from contextlib import nullcontext
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, TypeVar, cast

from . import background_tasks, core, helpers
from .awaitable_response import AwaitableResponse
from .slot import Slot

if TYPE_CHECKING:
    from .client import Client
    from .element import Element
    from .elements.slide_item import SlideSide
    from .elements.upload_files import FileUpload
    from .observables import ObservableCollection


@dataclass(kw_only=True, slots=True)
class EventArguments:
    pass


@dataclass(kw_only=True, slots=True)
class NativeEventArguments(EventArguments):
    type: str
    args: dict[str, Any]


@dataclass(kw_only=True, slots=True)
class ObservableChangeEventArguments(EventArguments):
    sender: ObservableCollection


@dataclass(kw_only=True, slots=True)
class UiEventArguments(EventArguments):
    sender: Element
    client: Client


@dataclass(kw_only=True, slots=True)
class GenericEventArguments(UiEventArguments):
    args: Any


@dataclass(kw_only=True, slots=True)
class ClickEventArguments(UiEventArguments):
    pass


@dataclass(kw_only=True, slots=True)
class SlideEventArguments(UiEventArguments):
    side: SlideSide


@dataclass(kw_only=True, slots=True)
class EChartComponentClickEventArguments(UiEventArguments):
    component_type: str
    name: str | None


@dataclass(kw_only=True, slots=True)
class EChartPointClickEventArguments(UiEventArguments):
    component_type: str
    name: str
    series_type: str
    series_index: int
    series_name: str
    data_index: int
    data: float | int | str
    data_type: str | None
    value: float | int | list


@dataclass(kw_only=True, slots=True)
class MermaidNodeClickEventArguments(UiEventArguments):
    node_id: str


@dataclass(kw_only=True, slots=True)
class SceneClickHit:
    object_id: str
    object_name: str
    x: float
    y: float
    z: float


@dataclass(kw_only=True, slots=True)
class SceneClickEventArguments(ClickEventArguments):
    click_type: str
    button: int
    alt: bool
    ctrl: bool
    meta: bool
    shift: bool
    hits: list[SceneClickHit]


@dataclass(kw_only=True, slots=True)
class SceneDragEventArguments(ClickEventArguments):
    type: Literal['dragstart', 'dragend']
    object_id: str
    object_name: str
    x: float
    y: float
    z: float


@dataclass(kw_only=True, slots=True)
class ColorPickEventArguments(UiEventArguments):
    color: str


@dataclass(kw_only=True, slots=True)
class MouseEventArguments(UiEventArguments):
    type: str
    image_x: float
    image_y: float
    button: int
    buttons: int
    alt: bool
    ctrl: bool
    meta: bool
    shift: bool


@dataclass(kw_only=True, slots=True)
class JoystickEventArguments(UiEventArguments):
    action: str
    x: float | None = None
    y: float | None = None


@dataclass(kw_only=True, slots=True)
class UploadEventArguments(UiEventArguments):
    file: FileUpload


@dataclass(kw_only=True, slots=True)
class MultiUploadEventArguments(UiEventArguments):
    files: list[FileUpload]


@dataclass(kw_only=True, slots=True)
class ValueChangeEventArguments(UiEventArguments):
    value: Any
    previous_value: Any = ...

    def __post_init__(self):
        # DEPRECATED: previous_value will be required in NiceGUI 4.0
        if self.previous_value is ...:
            helpers.warn_once('The new event argument `ValueChangeEventArguments.previous_value` is not set. '
                              'In NiceGUI 4.0 this will raise an error.')


@dataclass(kw_only=True, slots=True)
class TableSelectionEventArguments(UiEventArguments):
    selection: list[Any]


@dataclass(kw_only=True, slots=True)
class KeyboardAction:
    keydown: bool
    keyup: bool
    repeat: bool


@dataclass(kw_only=True, slots=True)
class KeyboardModifiers:
    alt: bool
    ctrl: bool
    meta: bool
    shift: bool

    def __iter__(self) -> Iterator[bool]:
        return iter([self.alt, self.ctrl, self.meta, self.shift])

    def __len__(self) -> int:
        return sum(self)


@dataclass(kw_only=True, slots=True)
class KeyboardKey:
    name: str
    code: str
    location: int

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return other in {self.name, self.code}
        if isinstance(other, KeyboardKey):
            return (self.name, self.code, self.location) == (other.name, other.code, other.location)
        return False

    def __hash__(self) -> int:
        return hash((self.name, self.code, self.location))

    def __repr__(self):
        return str(self.name)

    @property
    def is_cursorkey(self) -> bool:
        """Whether the key is a cursor key (arrow keys)."""
        return self.code.startswith('Arrow')

    @property
    def number(self) -> int | None:
        """Integer value of a number key."""
        return int(self.code[len('Digit'):]) if self.code.startswith('Digit') else None

    @property
    def backspace(self) -> bool:
        """Whether the key is the backspace key."""
        return self.name == 'Backspace'

    @property
    def tab(self) -> bool:
        """Whether the key is the tab key."""
        return self.name == 'Tab'

    @property
    def enter(self) -> bool:
        """Whether the key is the enter key."""
        return self.name == 'Enter'

    @property
    def shift(self) -> bool:
        """Whether the key is the shift key."""
        return self.name == 'Shift'

    @property
    def control(self) -> bool:
        """Whether the key is the control key."""
        return self.name == 'Control'

    @property
    def alt(self) -> bool:
        """Whether the key is the alt key."""
        return self.name == 'Alt'

    @property
    def pause(self) -> bool:
        """Whether the key is the pause key."""
        return self.name == 'Pause'

    @property
    def caps_lock(self) -> bool:
        """Whether the key is the caps lock key."""
        return self.name == 'CapsLock'

    @property
    def escape(self) -> bool:
        """Whether the key is the escape key."""
        return self.name == 'Escape'

    @property
    def space(self) -> bool:
        """Whether the key is the space key."""
        return self.name == ' '

    @property
    def page_up(self) -> bool:
        """Whether the key is the page up key."""
        return self.name == 'PageUp'

    @property
    def page_down(self) -> bool:
        """Whether the key is the page down key."""
        return self.name == 'PageDown'

    @property
    def end(self) -> bool:
        """Whether the key is the end key."""
        return self.name == 'End'

    @property
    def home(self) -> bool:
        """Whether the key is the home key."""
        return self.name == 'Home'

    @property
    def arrow_left(self) -> bool:
        """Whether the key is the arrow left key."""
        return self.name == 'ArrowLeft'

    @property
    def arrow_up(self) -> bool:
        """Whether the key is the arrow up key."""
        return self.name == 'ArrowUp'

    @property
    def arrow_right(self) -> bool:
        """Whether the key is the arrow right key."""
        return self.name == 'ArrowRight'

    @property
    def arrow_down(self) -> bool:
        """Whether the key is the arrow down key."""
        return self.name == 'ArrowDown'

    @property
    def print_screen(self) -> bool:
        """Whether the key is the print screen key."""
        return self.name == 'PrintScreen'

    @property
    def insert(self) -> bool:
        """Whether the key is the insert key."""
        return self.name == 'Insert'

    @property
    def delete(self) -> bool:
        """Whether the key is the delete key."""
        return self.name == 'Delete'

    @property
    def meta(self) -> bool:
        """Whether the key is the meta key."""
        return self.name == 'Meta'

    @property
    def f1(self) -> bool:
        """Whether the key is the F1 key."""
        return self.name == 'F1'

    @property
    def f2(self) -> bool:
        """Whether the key is the F2 key."""
        return self.name == 'F2'

    @property
    def f3(self) -> bool:
        """Whether the key is the F3 key."""
        return self.name == 'F3'

    @property
    def f4(self) -> bool:
        """Whether the key is the F4 key."""
        return self.name == 'F4'

    @property
    def f5(self) -> bool:
        """Whether the key is the F5 key."""
        return self.name == 'F5'

    @property
    def f6(self) -> bool:
        """Whether the key is the F6 key."""
        return self.name == 'F6'

    @property
    def f7(self) -> bool:
        """Whether the key is the F7 key."""
        return self.name == 'F7'

    @property
    def f8(self) -> bool:
        """Whether the key is the F8 key."""
        return self.name == 'F8'

    @property
    def f9(self) -> bool:
        """Whether the key is the F9 key."""
        return self.name == 'F9'

    @property
    def f10(self) -> bool:
        """Whether the key is the F10 key."""
        return self.name == 'F10'

    @property
    def f11(self) -> bool:
        """Whether the key is the F11 key."""
        return self.name == 'F11'

    @property
    def f12(self) -> bool:
        """Whether the key is the F12 key."""
        return self.name == 'F12'


@dataclass(kw_only=True, slots=True)
class KeyEventArguments(UiEventArguments):
    action: KeyboardAction
    key: KeyboardKey
    modifiers: KeyboardModifiers


@dataclass(kw_only=True, slots=True)
class ScrollEventArguments(UiEventArguments):
    vertical_position: float
    vertical_percentage: float
    vertical_size: float
    vertical_container_size: float
    horizontal_position: float
    horizontal_percentage: float
    horizontal_size: float
    horizontal_container_size: float


@dataclass(kw_only=True, slots=True)
class JsonEditorSelectEventArguments(UiEventArguments):
    selection: dict


@dataclass(kw_only=True, slots=True)
class JsonEditorChangeEventArguments(UiEventArguments):
    content: dict
    errors: dict = field(default_factory=dict)


@dataclass(kw_only=True, slots=True)
class XtermBellEventArguments(UiEventArguments):
    pass


@dataclass(kw_only=True, slots=True)
class XtermDataEventArguments(UiEventArguments):
    data: str


EventT = TypeVar('EventT', bound=EventArguments)
Handler: TypeAlias = Callable[[EventT], Any] | Callable[[], Any]


def handle_event(handler: Handler[EventT] | None, arguments: EventT) -> None:
    """Call the given event handler.

    The handler is called within the context of the parent slot of the sender.
    If the handler is a coroutine, it is scheduled as a background task.
    If the handler expects arguments, the arguments are passed to the handler.
    Exceptions are caught and handled globally.

    :param handler: the event handler
    :param arguments: the event arguments
    """
    if handler is None:
        return
    try:
        parent_slot: Slot | nullcontext
        if isinstance(arguments, UiEventArguments):
            parent_slot = arguments.sender.parent_slot or arguments.sender.client.layout.default_slot
        else:
            parent_slot = nullcontext()

        with parent_slot:
            if helpers.expects_arguments(handler):
                result = cast(Callable[[EventT], Any], handler)(arguments)
            else:
                result = cast(Callable[[], Any], handler)()
        if isinstance(result, Awaitable) and not isinstance(result, AwaitableResponse) and not isinstance(result, asyncio.Task):
            # NOTE: await an awaitable result even if the handler is not a coroutine (like a lambda statement)
            async def wait_for_result():
                with parent_slot:
                    try:
                        await result
                    except Exception as e:
                        core.app.handle_exception(e)
            if core.loop and core.loop.is_running():
                background_tasks.create(wait_for_result(), name=str(handler))
            else:
                core.app.on_startup(wait_for_result())
    except Exception as e:
        core.app.handle_exception(e)



from pathlib import Path

import pytest

from . import general

# pylint: disable=redefined-outer-name


def pytest_addoption(parser: pytest.Parser) -> None:
    """Add pytest option for main file."""
    parser.addini('main_file', 'main file', default='main.py')


def pytest_configure(config: pytest.Config) -> None:
    """Register the "nicegui_main_file" marker."""
    config.addinivalue_line('markers', 'nicegui_main_file: specify the main file for the test')


def get_path_to_main_file(request: pytest.FixtureRequest) -> Path | None:
    """Get the path to the main file from the test marker or global config."""
    marker = next((m for m in request.node.iter_markers('nicegui_main_file')), None)
    main_file = marker.args[0] if marker else request.config.getini('main_file')
    if not main_file:
        return None
    assert request.config.inipath is not None
    path = (request.config.inipath.parent / main_file).resolve()
    if not path.is_file():
        raise FileNotFoundError(f'Main file not found: {path}')
    return path


@pytest.fixture
def nicegui_reset_globals():
    """Reset the global state of the NiceGUI package."""
    with general.nicegui_reset_globals():
        yield



from collections.abc import Callable
from typing import Any, Literal, cast

from typing_extensions import Self

from nicegui.binding import BindableProperty, bind, bind_from, bind_to

from ...element import Element

QUASAR_COLORS = {'primary', 'secondary', 'accent', 'dark', 'positive', 'negative', 'info', 'warning'}
for color in ['red', 'pink', 'purple', 'deep-purple', 'indigo', 'blue', 'light-blue', 'cyan', 'teal', 'green',
              'light-green', 'lime', 'yellow', 'amber', 'orange', 'deep-orange', 'brown', 'grey', 'blue-grey']:
    QUASAR_COLORS.add(color)
    for i in range(1, 15):
        QUASAR_COLORS.add(f'{color}-{i}')

TAILWIND_COLORS = {
    f'{name}-{value}'
    for name in ['red', 'orange', 'amber', 'yellow', 'lime', 'green', 'emerald', 'teal', 'cyan', 'sky', 'blue',
                 'indigo', 'violet', 'purple', 'fuchsia', 'pink', 'rose', 'slate', 'gray', 'zinc', 'neutral', 'stone']
    for value in [50, 100, 200, 300, 400, 500, 600, 700, 800, 900, 950]
}


class BackgroundColorElement(Element):
    BACKGROUND_COLOR_PROP = 'color'
    background_color = BindableProperty(
        on_change=lambda sender, background_color: cast(Self, sender)._handle_background_color_change(background_color))  # pylint: disable=protected-access

    def __init__(self, *, background_color: str | None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._background_color_state: tuple[Literal[None, 'prop', 'class', 'style'], str] = (None, '')
        self.background_color = background_color
        self._handle_background_color_change(background_color)

    def set_background_color(self, background_color: str | None) -> None:
        """Set the background color of this element."""
        self.background_color = background_color

    def _handle_background_color_change(self, background_color: str | None) -> None:
        # Clear previous color based on tracked state
        color_type, value = self._background_color_state
        if color_type == 'prop':
            self._props.pop(self.BACKGROUND_COLOR_PROP, None)
        elif color_type == 'class':
            self._classes.remove(f'bg-{value}')
        elif color_type == 'style':
            self._style.pop('background-color', None)

        # Set new color and track state
        if background_color in QUASAR_COLORS or background_color is None:
            self._props.set_optional(self.BACKGROUND_COLOR_PROP, background_color)
            self._background_color_state = ('prop', background_color) if background_color else (None, '')
        elif background_color in TAILWIND_COLORS:
            self._classes.append(f'bg-{background_color}')
            self._background_color_state = ('class', background_color)
        else:
            self._style['background-color'] = background_color
            self._background_color_state = ('style', background_color)

    def bind_background_color_to(self,
                                 target_object: Any,
                                 target_name: str = 'background_color',
                                 forward: Callable[[Any], Any] | None = None, *,
                                 strict: bool | None = None,
                                 ) -> Self:
        """Bind the background color of this element to the target object's target_name property.

        The binding works one way only, from this element to the target.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_to(self, 'background_color', target_object, target_name, forward, self_strict=False, other_strict=strict)
        return self

    def bind_background_color_from(self,
                                   target_object: Any,
                                   target_name: str = 'background_color',
                                   backward: Callable[[Any], Any] | None = None, *,
                                   strict: bool | None = None,
                                   ) -> Self:
        """Bind the background color of this element from the target object's target_name property.

        The binding works one way only, from the target to this element.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind from.
        :param target_name: The name of the property to bind from.
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_from(self, 'background_color', target_object, target_name,
                  backward, self_strict=False, other_strict=strict)
        return self

    def bind_background_color(self,
                              target_object: Any,
                              target_name: str = 'background_color', *,
                              forward: Callable[[Any], Any] | None = None,
                              backward: Callable[[Any], Any] | None = None,
                              strict: bool | None = None,
                              ) -> Self:
        """Bind the background color of this element to the target object's target_name property.

        The binding works both ways, from this element to the target and from the target to this element.
        The update happens immediately and whenever a value changes.
        The backward binding takes precedence for the initial synchronization.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind(self, 'background_color', target_object, target_name,
             forward=forward, backward=backward,
             self_strict=False, other_strict=strict)
        return self


class TextColorElement(Element):
    TEXT_COLOR_PROP = 'color'

    text_color = BindableProperty(
        on_change=lambda sender, text_color: cast(Self, sender)._handle_text_color_change(text_color))  # pylint: disable=protected-access

    def __init__(self, *, text_color: str | None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._text_color_state: tuple[Literal[None, 'prop', 'class', 'style'], str] = (None, '')
        self.text_color = text_color
        self._handle_text_color_change(text_color)

    def set_text_color(self, text_color: str | None) -> None:
        """Set the text color of this element."""
        self.text_color = text_color

    def _handle_text_color_change(self, text_color: str | None) -> None:
        # Clear previous color based on tracked state
        color_type, value = self._text_color_state
        if color_type == 'prop':
            self._props.pop(self.TEXT_COLOR_PROP, None)
        elif color_type == 'class':
            self._classes.remove(f'text-{value}')
        elif color_type == 'style':
            self._style.pop('color', None)

        # Set new color and track state
        if text_color in QUASAR_COLORS or text_color is None:
            self._props.set_optional(self.TEXT_COLOR_PROP, text_color)
            self._text_color_state = ('prop', text_color) if text_color else (None, '')
        elif text_color in TAILWIND_COLORS:
            self._classes.append(f'text-{text_color}')
            self._text_color_state = ('class', text_color)
        else:
            self._style['color'] = text_color
            self._text_color_state = ('style', text_color)

    def bind_text_color_to(self,
                           target_object: Any,
                           target_name: str = 'text_color',
                           forward: Callable[[Any], Any] | None = None, *,
                           strict: bool | None = None,
                           ) -> Self:
        """Bind the text color of this element to the target object's target_name property.

        The binding works one way only, from this element to the target.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_to(self, 'text_color', target_object, target_name, forward, self_strict=False, other_strict=strict)
        return self

    def bind_text_color_from(self,
                             target_object: Any,
                             target_name: str = 'text_color',
                             backward: Callable[[Any], Any] | None = None, *,
                             strict: bool | None = None,
                             ) -> Self:
        """Bind the text color of this element from the target object's target_name property.

        The binding works one way only, from the target to this element.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind from.
        :param target_name: The name of the property to bind from.
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_from(self, 'text_color', target_object, target_name,
                  backward, self_strict=False, other_strict=strict)
        return self

    def bind_text_color(self,
                        target_object: Any,
                        target_name: str = 'text_color', *,
                        forward: Callable[[Any], Any] | None = None,
                        backward: Callable[[Any], Any] | None = None,
                        strict: bool | None = None,
                        ) -> Self:
        """Bind the text color of this element to the target object's target_name property.

        The binding works both ways, from this element to the target and from the target to this element.
        The update happens immediately and whenever a value changes.
        The backward binding takes precedence for the initial synchronization.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
              performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind(self, 'text_color', target_object, target_name,
             forward=forward, backward=backward,
             self_strict=False, other_strict=strict)
        return self



from __future__ import annotations

import tempfile
import time
from contextlib import suppress
from pathlib import Path

from .. import optional_features
from ..logging import log
from .mixins.source_element import SourceElement

with suppress(ImportError):
    from PIL.Image import Image as PIL_Image
    optional_features.register('pillow')


class Image(SourceElement, component='image.js'):
    PIL_CONVERT_FORMAT = 'PNG'

    def __init__(self, source: str | Path | PIL_Image = '') -> None:
        """Image

        Displays an image.
        This element is based on Quasar's `QImg <https://quasar.dev/vue-components/img>`_ component.

        :param source: the source of the image; can be a URL, local file path, a base64 string or a PIL image
        """
        super().__init__(source=source)

    def set_source(self, source: str | Path | PIL_Image) -> None:
        return super().set_source(source)

    def _set_props(self, source: str | Path | PIL_Image) -> None:
        if optional_features.has('pillow') and isinstance(source, PIL_Image):
            source = pil_to_tempfile(source, self.PIL_CONVERT_FORMAT)
        super()._set_props(source)

    def force_reload(self) -> None:
        """Force the image to reload from the source."""
        if self._props['src'].startswith('data:'):
            log.warning('ui.image: force_reload() only works with network sources (not data URIs)')
            return
        self._props['t'] = time.time()


def pil_to_tempfile(pil_image: PIL_Image, image_format: str) -> _TempPath:
    """Save a PIL image to a temporary file.

    :param pil_image: the PIL image
    :param image_format: the image format
    :return: the path to the temporary file (auto-deletes when dereferenced)
    """
    suffix = f'.{image_format.lower()}'
    with tempfile.NamedTemporaryFile(delete=False, suffix=suffix) as temp_file:
        pil_image.save(temp_file, image_format)
        return _TempPath(temp_file.name)


class _TempPath(type(Path())):  # type: ignore[misc]  # NOTE: Path is not subclassable before Python 3.12
    """A Path subclass that deletes itself when garbage collected."""

    def __del__(self) -> None:
        self.unlink(missing_ok=True)
